from netqasm.lang.operand import Template

from qoala.lang.parse import RequestRoutineParser
from qoala.lang.request import RequestVirtIdMapping, VirtIdMappingType


def test_virt_id_mapping_to_string():
    all_zero = RequestVirtIdMapping(
        typ=VirtIdMappingType.EQUAL, single_value=0, custom_values=None
    )
    assert str(all_zero) == "all 0"

    all_three = RequestVirtIdMapping(
        typ=VirtIdMappingType.EQUAL, single_value=3, custom_values=None
    )
    assert str(all_three) == "all 3"

    increment_one = RequestVirtIdMapping(
        typ=VirtIdMappingType.INCREMENT, single_value=1, custom_values=None
    )
    assert str(increment_one) == "increment 1"

    custom = RequestVirtIdMapping(
        typ=VirtIdMappingType.CUSTOM, single_value=None, custom_values=[1, 2, 5]
    )
    assert str(custom) == "custom 1, 2, 5"

    all_template = RequestVirtIdMapping(
        typ=VirtIdMappingType.EQUAL,
        single_value=Template("virt_id"),
        custom_values=None,
    )
    assert str(all_template) == "all {virt_id}"

    increment_template = RequestVirtIdMapping(
        typ=VirtIdMappingType.INCREMENT,
        single_value=Template("virt_id"),
        custom_values=None,
    )
    assert str(increment_template) == "increment {virt_id}"


def test_string_to_virt_id_mapping():
    all_zero = RequestVirtIdMapping(
        typ=VirtIdMappingType.EQUAL, single_value=0, custom_values=None
    )
    assert RequestVirtIdMapping.from_str("all 0") == all_zero

    all_three = RequestVirtIdMapping(
        typ=VirtIdMappingType.EQUAL, single_value=3, custom_values=None
    )
    assert RequestVirtIdMapping.from_str("all 3") == all_three

    increment_one = RequestVirtIdMapping(
        typ=VirtIdMappingType.INCREMENT, single_value=1, custom_values=None
    )
    assert RequestVirtIdMapping.from_str("increment 1") == increment_one

    custom = RequestVirtIdMapping(
        typ=VirtIdMappingType.CUSTOM, single_value=None, custom_values=[1, 2, 5]
    )
    assert RequestVirtIdMapping.from_str("custom 1, 2, 5") == custom

    all_template = RequestVirtIdMapping(
        typ=VirtIdMappingType.EQUAL,
        single_value=Template("virt_id"),
        custom_values=None,
    )
    assert RequestVirtIdMapping.from_str("all {virt_id}") == all_template

    increment_template = RequestVirtIdMapping(
        typ=VirtIdMappingType.INCREMENT,
        single_value=Template("virt_id"),
        custom_values=None,
    )
    assert RequestVirtIdMapping.from_str("increment {virt_id}") == increment_template


def test_instantiate():
    text = """
REQUEST req1
  callback_type: wait_all
  callback: 
  return_vars: 
  remote_id: {client_id}
  epr_socket_id: 0
  num_pairs: {num_pairs}
  virt_ids: custom 1, 2, 3
  timeout: 1000
  fidelity: 0.65
  typ: measure_directly
  role: receive
    """

    request_routine = RequestRoutineParser(text).parse()["req1"]
    request = request_routine.request
    assert request.remote_id == Template(name="client_id")
    assert request.num_pairs == Template(name="num_pairs")

    request.instantiate(values={"client_id": 2, "num_pairs": 10})
    assert request.remote_id == 2
    assert request.num_pairs == 10


if __name__ == "__main__":
    test_virt_id_mapping_to_string()
    test_string_to_virt_id_mapping()
    test_instantiate()



from __future__ import annotations

import os
from typing import List

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.program import BatchInfo, ProgramInput
from qoala.runtime.schedule import TaskSchedule
from qoala.sim.build import build_network
from qoala.sim.network import ProcNodeNetwork


def create_network_info() -> NetworkInfo:
    env = NetworkInfo.with_nodes({0: "alice"})

    env.set_global_schedule([0])
    env.set_timeslot_len(1e6)

    return env


def get_config() -> ProcNodeConfig:
    topology = TopologyConfig.perfect_config_uniform_default_params(1)
    return ProcNodeConfig(
        node_name="alice",
        node_id=0,
        topology=topology,
        latencies=LatenciesConfig(qnos_instr_time=1000),
    )


def create_network(
    node_cfg: ProcNodeConfig,
) -> ProcNodeNetwork:
    network_info = create_network_info()

    network_cfg = ProcNodeNetworkConfig(nodes=[node_cfg], links=[])
    return build_network(network_cfg, network_info)


def load_program() -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), "simple_program.iqoala")
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text).parse()

    return program


def create_batch(
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
    deadline: int,
) -> BatchInfo:
    program = load_program()

    return BatchInfo(
        program=program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=deadline,
    )


def run_program():
    ns.sim_reset()

    node_config = get_config()
    network = create_network(node_config)
    procnode = network.nodes["alice"]

    num_iterations = 100
    inputs = [ProgramInput({}) for i in range(num_iterations)]

    unit_module = UnitModule.from_full_ehi(procnode.memmgr.get_ehi())

    batch_info = create_batch(
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )

    procnode.submit_batch(batch_info)
    procnode.initialize_processes()
    tasks = procnode.scheduler.get_tasks_to_schedule()
    schedule = TaskSchedule.consecutive(tasks)
    procnode.scheduler.upload_schedule(schedule)

    network.start_all_nodes()
    ns.sim_run()

    all_results = procnode.scheduler.get_batch_results()
    batch0_result = all_results[0]
    results = [result.values["m"] for result in batch0_result.results]
    print(results)


def test_simple_program():
    # LogManager.set_log_level("DEBUG")
    # LogManager.log_to_file("logs/simple_program.log")

    run_program()


if __name__ == "__main__":
    test_simple_program()



from __future__ import annotations

from typing import Generator

from pydynaa import EventExpression
from qoala.runtime.message import Message
from qoala.sim.host.hostinterface import HostInterface


class ClassicalSocket:
    """Wrapper around classical ports"""

    def __init__(self, host: HostInterface, remote_name: str):
        self._host = host
        self._remote_name = remote_name

    def send(self, msg: Message) -> None:
        """Send a message to the remote node."""
        self._host.send_peer_msg(self._remote_name, msg)

    def recv(self) -> Generator[EventExpression, None, Message]:
        msg = yield from self._host.receive_peer_msg(self._remote_name)
        return msg

    def send_str(self, msg: str) -> None:
        self.send(Message(content=msg))

    def recv_str(self) -> Generator[EventExpression, None, str]:
        msg = yield from self.recv()
        assert isinstance(msg.content, str)
        return msg.content

    def send_int(self, value: int) -> None:
        self.send(Message(content=value))

    def recv_int(self) -> Generator[EventExpression, None, int]:
        msg = yield from self.recv()
        assert isinstance(msg.content, int)
        return msg.content

    def send_float(self, value: float) -> None:
        self.send(Message(content=value))

    def recv_float(self) -> Generator[EventExpression, None, float]:
        msg = yield from self.recv()
        assert isinstance(msg.content, float)
        return msg.content



from __future__ import annotations

import itertools
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Type

from netqasm.lang.instr.base import NetQASMInstruction
from netqasm.lang.instr.flavour import Flavour

from qoala.lang.common import MultiQubit


@dataclass(eq=True, frozen=True)
class EhiQubitInfo:
    is_communication: bool
    decoherence_rate: float  # rate per second


@dataclass(eq=True, frozen=True)
class EhiGateInfo:
    instruction: Type[NetQASMInstruction]
    duration: float  # ns
    decoherence: float  # rate per second, for all qubits


@dataclass(eq=True, frozen=True)
class EhiLinkInfo:
    duration: float  # ns
    fidelity: float


@dataclass(eq=True, frozen=True)
class EhiLatencies:
    host_instr_time: float  # duration of classical Host instr execution (CL)
    qnos_instr_time: float  # duration of classical Qnos instr execution (QL)
    host_peer_latency: float  # processing time for Host messages from remote node (CC)

    @classmethod
    def all_zero(cls) -> EhiLatencies:
        return EhiLatencies(0, 0, 0)


@dataclass(eq=True, frozen=True)
class EhiNodeInfo:
    """Hardware made available to offline compiler."""

    qubit_infos: Dict[int, EhiQubitInfo]  # qubit ID -> info

    flavour: Type[
        Flavour
    ]  # set of NetQASM instrs, no info about which qubits can do what instr
    single_gate_infos: Dict[int, List[EhiGateInfo]]  # qubit ID -> gates
    multi_gate_infos: Dict[
        MultiQubit, List[EhiGateInfo]
    ]  # ordered qubit ID list -> gates

    latencies: EhiLatencies

    def find_single_gate(
        self, qubit_id: int, instr: Type[NetQASMInstruction]
    ) -> Optional[EhiGateInfo]:
        if qubit_id not in self.single_gate_infos:
            return None
        for info in self.single_gate_infos[qubit_id]:
            if info.instruction == instr:
                return info
        return None

    def find_multi_gate(
        self, qubit_ids: List[int], instr: Type[NetQASMInstruction]
    ) -> Optional[EhiGateInfo]:
        multi = MultiQubit(qubit_ids)
        if multi not in self.multi_gate_infos:
            return None
        for info in self.multi_gate_infos[multi]:
            if info.instruction == instr:
                return info
        return None


class EhiBuilder:
    @classmethod
    def decoherence_qubit(
        cls, is_communication: bool, decoherence_rate: float
    ) -> EhiQubitInfo:
        return EhiQubitInfo(
            is_communication=is_communication, decoherence_rate=decoherence_rate
        )

    @classmethod
    def perfect_qubit(cls, is_communication: bool) -> EhiQubitInfo:
        return cls.decoherence_qubit(
            is_communication=is_communication, decoherence_rate=0
        )

    @classmethod
    def decoherence_gates(
        cls,
        duration: float,
        instructions: List[Type[NetQASMInstruction]],
        decoherence: float,
    ) -> List[EhiGateInfo]:
        return [
            EhiGateInfo(instruction=instr, duration=duration, decoherence=decoherence)
            for instr in instructions
        ]

    @classmethod
    def perfect_gates(
        cls, duration: float, instructions: List[Type[NetQASMInstruction]]
    ) -> List[EhiGateInfo]:
        return cls.decoherence_gates(
            duration=duration, instructions=instructions, decoherence=0
        )

    @classmethod
    def perfect_uniform(
        cls,
        num_qubits,
        flavour: Type[Flavour],
        single_instructions: List[Type[NetQASMInstruction]],
        single_duration: float,
        two_instructions: List[Type[NetQASMInstruction]],
        two_duration: float,
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        return cls.fully_uniform(
            num_qubits=num_qubits,
            flavour=flavour,
            qubit_info=cls.perfect_qubit(is_communication=True),
            single_gate_infos=cls.perfect_gates(single_duration, single_instructions),
            two_gate_infos=cls.perfect_gates(two_duration, two_instructions),
            latencies=latencies,
        )

    @classmethod
    def fully_uniform(
        cls,
        num_qubits,
        qubit_info: EhiQubitInfo,
        flavour: Type[Flavour],
        single_gate_infos: List[EhiGateInfo],
        two_gate_infos: List[EhiGateInfo],
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        q_infos = {i: qubit_info for i in range(num_qubits)}
        sg_infos = {i: single_gate_infos for i in range(num_qubits)}
        mg_infos = {}
        for i in range(num_qubits):
            for j in range(num_qubits):
                if i != j:
                    multi = MultiQubit([i, j])
                    mg_infos[multi] = two_gate_infos

        if latencies is None:
            latencies = EhiLatencies.all_zero()
        return EhiNodeInfo(q_infos, flavour, sg_infos, mg_infos, latencies)

    @classmethod
    def perfect_star(
        cls,
        num_qubits: int,
        flavour: Type[Flavour],
        comm_instructions: List[Type[NetQASMInstruction]],
        comm_duration: float,
        mem_instructions: List[Type[NetQASMInstruction]],
        mem_duration: float,
        two_instructions: List[Type[NetQASMInstruction]],
        two_duration: float,
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        comm_qubit_info = cls.perfect_qubit(is_communication=True)
        mem_qubit_info = cls.perfect_qubit(is_communication=False)
        comm_gate_infos = cls.perfect_gates(comm_duration, comm_instructions)
        mem_gate_infos = cls.perfect_gates(mem_duration, mem_instructions)
        two_gate_infos = cls.perfect_gates(two_duration, two_instructions)

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        if latencies is None:
            latencies = EhiLatencies.all_zero()
        return EhiNodeInfo(q_infos, flavour, sg_infos, mg_infos, latencies)

    @classmethod
    def generic_t1t2_star(
        cls,
        num_qubits: int,
        flavour: Type[Flavour],
        comm_decoherence: float,
        mem_decoherence: float,
        comm_instructions: List[Type[NetQASMInstruction]],
        comm_duration: float,
        comm_instr_decoherence: float,
        mem_instructions: List[Type[NetQASMInstruction]],
        mem_duration: float,
        mem_instr_decoherence: float,
        two_instructions: List[Type[NetQASMInstruction]],
        two_duration: float,
        two_instr_decoherence: float,
        latencies: Optional[EhiLatencies] = None,
    ) -> EhiNodeInfo:
        comm_qubit_info = cls.decoherence_qubit(
            is_communication=True, decoherence_rate=comm_decoherence
        )
        mem_qubit_info = cls.decoherence_qubit(
            is_communication=False, decoherence_rate=mem_decoherence
        )
        comm_gate_infos = cls.decoherence_gates(
            comm_duration, comm_instructions, comm_instr_decoherence
        )
        mem_gate_infos = cls.decoherence_gates(
            mem_duration, mem_instructions, mem_instr_decoherence
        )
        two_gate_infos = cls.decoherence_gates(
            two_duration, two_instructions, two_instr_decoherence
        )

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        if latencies is None:
            latencies = EhiLatencies.all_zero()
        return EhiNodeInfo(q_infos, flavour, sg_infos, mg_infos, latencies)


@dataclass(eq=True, frozen=True)
class UnitModule:
    """Description of virtual memory space for programs. Target for a compiler.

    Simply wraps around a EhiNodeInfo object and provides convenience methods.

    Unit Modules should be used as the interface for compilers and schedulers,
    as well as the program itself. This object does not contain information about
    runtime values (i.e. qubit mappings); this is managed by the Memory Manager.
    Only the Memory Manager should use an EhiNodeInfo object itself,
    namely the object that represents the full quantum memory space of the node."""

    info: EhiNodeInfo

    def is_communication(self, qubit_id: int) -> bool:
        return self.info.qubit_infos[qubit_id].is_communication

    @classmethod
    def from_ehi(cls, ehi: EhiNodeInfo, qubit_ids: List[int]) -> UnitModule:
        """Get a subset of an EhiNodeInfo"""
        qubit_infos = {i: ehi.qubit_infos[i] for i in qubit_ids}
        single_gate_infos = {i: ehi.single_gate_infos[i] for i in qubit_ids}
        multi_gate_infos = {
            ids: info
            for (ids, info) in ehi.multi_gate_infos.items()
            if all(id in qubit_ids for id in ids.qubit_ids)
        }

        return UnitModule(
            info=EhiNodeInfo(
                qubit_infos,
                ehi.flavour,
                single_gate_infos,
                multi_gate_infos,
                ehi.latencies,
            )
        )

    @classmethod
    def from_full_ehi(cls, ehi: EhiNodeInfo) -> UnitModule:
        """Use the full EhiNodeInfo"""
        qubit_ids = [i for i in ehi.qubit_infos.keys()]
        return cls.from_ehi(ehi, qubit_ids)

    def get_all_qubit_ids(self) -> List[int]:
        return list(self.info.qubit_infos.keys())


@dataclass
class EhiNetworkInfo:
    # (node A ID, node B ID) -> link info
    # for a pair (a, b) there exists no separate (b, a) info (it is the same)
    links: Dict[Tuple[int, int], EhiLinkInfo]

    @classmethod
    def fully_connected(cls, node_ids: List[int], info: EhiLinkInfo) -> EhiNetworkInfo:
        links: Dict[Tuple[int, int], EhiLinkInfo] = {}
        for n1, n2 in itertools.combinations(node_ids, 2):
            links[(n1, n2)] = info
        return EhiNetworkInfo(links)

    @classmethod
    def perfect_fully_connected(
        cls, node_ids: List[int], duration: float
    ) -> EhiNetworkInfo:
        link = EhiLinkInfo(duration=duration, fidelity=1.0)
        return cls.fully_connected(node_ids, link)

    def get_link(self, node_id1: int, node_id2: int) -> EhiLinkInfo:
        try:
            return self.links[(node_id1, node_id2)]
        except KeyError:
            return self.links[(node_id2, node_id1)]



from __future__ import annotations

from netsquid.protocols import Protocol

from qoala.runtime.environment import LocalEnvironment
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.hostcomp import HostComponent
from qoala.sim.host.hostinterface import HostInterface, HostLatencies
from qoala.sim.host.hostprocessor import HostProcessor


class Host(Protocol):
    """NetSquid protocol representing a Host."""

    def __init__(
        self,
        comp: HostComponent,
        local_env: LocalEnvironment,
        latencies: HostLatencies,
        asynchronous: bool = False,
    ) -> None:
        """Host protocol constructor.

        :param comp: NetSquid component representing the Host
        """
        super().__init__(name=f"{comp.name}_protocol")

        # References to objects.
        self._comp = comp
        self._local_env = local_env

        # Owned objects.
        self._interface = HostInterface(comp, local_env)
        self._processor = HostProcessor(self._interface, latencies, asynchronous)

    @property
    def interface(self) -> HostInterface:
        return self._interface

    @interface.setter
    def interface(self, interface: HostInterface) -> None:
        self._interface = interface
        self._processor._interface = interface

    @property
    def processor(self) -> HostProcessor:
        return self._processor

    @property
    def local_env(self) -> LocalEnvironment:
        return self._local_env

    def start(self) -> None:
        assert self._interface is not None
        super().start()
        self._interface.start()

    def stop(self) -> None:
        self._interface.stop()
        super().stop()

    def create_csocket(self, remote_name: str) -> ClassicalSocket:
        return ClassicalSocket(self._interface, remote_name)



from dataclasses import dataclass
from typing import Dict, List, Optional


class SharedMemWriteError(Exception):
    pass


class SharedMemReadError(Exception):
    pass


class SharedMemNotAllocatedError(Exception):
    pass


class SharedMemIllegalRegionError(Exception):
    pass


@dataclass(eq=True, frozen=True)
class MemAddr:
    addr: int


class NetQASMArrays:
    def __init__(self) -> None:
        self._next_addr: int = 0
        self._memory: Dict[MemAddr, List[Optional[int]]] = {}

    def allocate(self, addr: MemAddr, size: int) -> None:
        assert addr not in self._memory
        self._memory[addr] = [None] * size

    def write(self, addr: MemAddr, data: List[int], offset: int) -> None:
        if addr not in self._memory:
            raise SharedMemWriteError
        if len(self._memory[addr]) < offset + len(data):
            raise SharedMemWriteError
        for i in range(len(data)):
            self._memory[addr][i + offset] = data[i]

    def read(self, addr: MemAddr, size: int, offset: int) -> List[int]:
        if addr not in self._memory:
            raise SharedMemReadError
        if len(self._memory[addr]) < offset + size:
            raise SharedMemReadError
        array = self._memory[addr][offset : offset + size]
        assert all(v is not None for v in array)
        return array  # type: ignore


class SharedMemory:
    def __init__(self) -> None:
        self._arrays = NetQASMArrays()

        self._rr_in_addrs: List[MemAddr] = []
        self._rr_out_addrs: List[MemAddr] = []
        self._cr_in_addrs: List[MemAddr] = []
        self._lr_in_addrs: List[MemAddr] = []
        self._lr_out_addrs: List[MemAddr] = []

        self._addr_counter: int = 0

    def _allocate(self, size: int) -> MemAddr:
        addr = MemAddr(self._addr_counter)
        self._addr_counter += 1
        self._arrays.allocate(addr, size)
        return addr

    def allocate_rr_in(self, size: int) -> MemAddr:
        addr = self._allocate(size)
        self._rr_in_addrs.append(addr)
        return addr

    def write_rr_in(self, addr: MemAddr, data: List[int], offset: int = 0) -> None:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._rr_in_addrs:
            raise SharedMemIllegalRegionError
        self._arrays.write(addr, data, offset)

    def read_rr_in(self, addr: MemAddr, size: int, offset: int = 0) -> List[int]:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._rr_in_addrs:
            raise SharedMemIllegalRegionError
        data = self._arrays.read(addr, size, offset)
        return data

    def allocate_rr_out(self, size: int) -> MemAddr:
        addr = self._allocate(size)
        self._rr_out_addrs.append(addr)
        return addr

    def write_rr_out(self, addr: MemAddr, data: List[int], offset: int = 0) -> None:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._rr_out_addrs:
            raise SharedMemIllegalRegionError
        self._arrays.write(addr, data, offset)

    def read_rr_out(self, addr: MemAddr, size: int, offset: int = 0) -> List[int]:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._rr_out_addrs:
            raise SharedMemIllegalRegionError
        data = self._arrays.read(addr, size, offset)
        return data

    def allocate_cr_in(self, size: int) -> MemAddr:
        addr = self._allocate(size)
        self._cr_in_addrs.append(addr)
        return addr

    def write_cr_in(self, addr: MemAddr, data: List[int], offset: int = 0) -> None:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._cr_in_addrs:
            raise SharedMemIllegalRegionError
        self._arrays.write(addr, data, offset)

    def read_cr_in(self, addr: MemAddr, size: int, offset: int = 0) -> List[int]:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._cr_in_addrs:
            raise SharedMemIllegalRegionError
        data = self._arrays.read(addr, size, offset)
        return data

    def allocate_lr_in(self, size: int) -> MemAddr:
        addr = self._allocate(size)
        self._lr_in_addrs.append(addr)
        return addr

    def _check_allocated(self, addr: MemAddr) -> bool:
        # "hack": make use of fact that addr counter is never decreased
        return addr.addr < self._addr_counter

    def write_lr_in(self, addr: MemAddr, data: List[int], offset: int = 0) -> None:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._lr_in_addrs:
            raise SharedMemIllegalRegionError
        self._arrays.write(addr, data, offset)

    def read_lr_in(self, addr: MemAddr, size: int, offset: int = 0) -> List[int]:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._lr_in_addrs:
            raise SharedMemIllegalRegionError
        data = self._arrays.read(addr, size, offset)
        return data

    def allocate_lr_out(self, size: int) -> MemAddr:
        addr = self._allocate(size)
        self._lr_out_addrs.append(addr)
        return addr

    def write_lr_out(self, addr: MemAddr, data: List[int], offset: int = 0) -> None:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._lr_out_addrs:
            raise SharedMemIllegalRegionError
        self._arrays.write(addr, data, offset)

    def read_lr_out(self, addr: MemAddr, size: int, offset: int = 0) -> List[int]:
        if not self._check_allocated(addr):
            raise SharedMemNotAllocatedError
        if addr not in self._lr_out_addrs:
            raise SharedMemIllegalRegionError
        data = self._arrays.read(addr, size, offset)
        return data



from dataclasses import dataclass
from typing import Generator, List, Optional, Set

from netsquid.components.instructions import INSTR_INIT, Instruction
from netsquid.components.qprocessor import QuantumProcessor
from netsquid.components.qprogram import QuantumProgram
from netsquid.nodes import Node
from netsquid.qubits.qubit import Qubit

from pydynaa import EventExpression
from qoala.runtime.lhi import LhiTopology


class UnsupportedQDeviceCommandError(Exception):
    pass


class NonInitializedQubitError(Exception):
    pass


@dataclass(eq=True, frozen=True)
class QDeviceCommand:
    instr: Instruction
    indices: List[int]
    angle: Optional[float] = None


class QDevice:
    def __init__(self, node: Node, topology: LhiTopology) -> None:
        self._node = node
        self._topology = topology

    @property
    def qprocessor(self) -> QuantumProcessor:
        """Get the NetSquid `QuantumProcessor` object of this node."""
        return self._node.qmemory

    @property
    def node(self) -> Node:
        """Get the NetSquid `Node` object of this QDevice."""
        return self._node

    @property
    def topology(self) -> LhiTopology:
        return self._topology

    def get_qubit_count(self) -> int:
        return len(self.get_all_qubit_ids())

    def get_comm_qubit_count(self) -> int:
        return len(self.get_comm_qubit_ids())

    def get_non_comm_qubit_count(self) -> int:
        return len(self.get_non_comm_qubit_ids())

    def get_all_qubit_ids(self) -> Set[int]:
        return {q for q, _ in self.topology.qubit_infos.items()}

    def get_comm_qubit_ids(self) -> Set[int]:
        return {
            q for q, info in self.topology.qubit_infos.items() if info.is_communication
        }

    def get_non_comm_qubit_ids(self) -> Set[int]:
        return {
            q
            for q, info in self.topology.qubit_infos.items()
            if not info.is_communication
        }

    def is_allowed(self, cmd: QDeviceCommand) -> bool:
        all_phys_instructions = self.qprocessor.get_physical_instructions()

        # Get the physical instruction with the same type ('gate').
        matches = [
            i for i in all_phys_instructions if i.instruction.name == cmd.instr.name
        ]
        # Should be at least one matching.
        if len(matches) == 0:
            return False

        for phys_instr in matches:
            # If there is no topology, this instruction is allowed on any qubit.
            if phys_instr.topology is None:
                return True

            # Else, check if it is allowed for our current qubit(s).
            if len(cmd.indices) == 1:
                if cmd.indices[0] in phys_instr.topology:
                    return True
            else:
                if (cmd.indices[0], cmd.indices[1]) in phys_instr.topology:
                    return True

        # We didn't find any matching instruction.
        return False

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        self.qprocessor.mem_positions[id].in_use = in_use

    def execute_commands(
        self, commands: List[QDeviceCommand]
    ) -> Generator[EventExpression, None, Optional[int]]:
        """Can only return at most 1 measurement result."""
        prog = QuantumProgram()

        # TODO: rewrite this abomination

        for cmd in commands:
            # Check if this instruction is allowed on this processor.
            # If not, NetSquid will just silently skip this instruction which is confusing.
            if not self.is_allowed(cmd):
                raise UnsupportedQDeviceCommandError(cmd)

        for cmd in commands:
            # Check if the qubit has been initialized, since instructions won't work
            # if this is not the case.
            # Anything after an INSTR_INIT instruction is fine.
            # TODO: better logic for detecting INITs of individual qubits.
            if cmd.instr == INSTR_INIT:
                break
            for index in cmd.indices:
                if self.get_local_qubit(index) is None:
                    raise NonInitializedQubitError

        for cmd in commands:
            if cmd.angle is not None:
                prog.apply(cmd.instr, qubit_indices=cmd.indices, angle=cmd.angle)
            else:
                prog.apply(cmd.instr, qubit_indices=cmd.indices)
        yield self.qprocessor.execute_program(prog)

        last_result = prog.output["last"]
        if last_result is not None:
            meas_outcome: int = last_result[0]
            return meas_outcome
        return None

    def execute_program(
        self, prog: QuantumProgram
    ) -> Generator[EventExpression, None, None]:
        raise DeprecationWarning

    def get_local_qubit(self, index: int) -> Qubit:
        return self.qprocessor.peek([index])[0]

    def get_local_qubits(self, indices: List[int]) -> List[Qubit]:
        return self.qprocessor.peek(indices)  # type: ignore



from __future__ import annotations

from typing import Dict, Optional

from netsquid.components import QuantumProcessor
from netsquid.protocols import Protocol

from qoala.lang.ehi import EhiNetworkInfo, EhiNodeInfo
from qoala.runtime.environment import LocalEnvironment, NetworkInfo
from qoala.runtime.lhi import LhiLatencies, LhiTopology
from qoala.runtime.lhi_to_ehi import LhiConverter, NativeToFlavourInterface
from qoala.runtime.program import BatchInfo, ProgramBatch
from qoala.sim.host.host import Host
from qoala.sim.host.hostcomp import HostComponent
from qoala.sim.host.hostinterface import HostLatencies
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack import Netstack, NetstackComponent, NetstackLatencies
from qoala.sim.process import QoalaProcess
from qoala.sim.procnodecomp import ProcNodeComponent
from qoala.sim.qdevice import QDevice
from qoala.sim.qnos import Qnos, QnosComponent, QnosLatencies
from qoala.sim.scheduler import Scheduler


class ProcNode(Protocol):
    """NetSquid protocol representing a node with a software stack."""

    def __init__(
        self,
        name: str,
        network_info: NetworkInfo,
        qprocessor: QuantumProcessor,
        qdevice_topology: LhiTopology,
        latencies: LhiLatencies,
        ntf_interface: NativeToFlavourInterface,
        network_ehi: EhiNetworkInfo,
        node: Optional[ProcNodeComponent] = None,
        node_id: Optional[int] = None,
        scheduler: Optional[Scheduler] = None,
        asynchronous: bool = False,
    ) -> None:
        """ProcNode constructor.

        :param name: name of this node
        :param node: an existing ProcNodeComponent object containing the static
            components or None. If None, a ProcNodeComponent is automatically
            created.
        :param qdevice_type: hardware type of the QDevice, defaults to "generic"
        :param qprocessor: NetSquid `QuantumProcessor` representing the QDevice,
            defaults to None. If None, a QuantumProcessor is created
            automatically.
        :param node_id: ID to use for the internal NetSquid node object
        :param use_default_components: whether to automatically create NetSquid
            components for the Host and QNodeOS, defaults to True. If False,
            this allows for manually creating and adding these components.
        """
        super().__init__(name=f"{name}")
        if node:
            self._node = node
        else:
            self._node = ProcNodeComponent(name, qprocessor, network_info, node_id)

        self._network_info = network_info
        self._network_ehi = network_ehi
        self._local_env = LocalEnvironment(network_info, network_info.get_node_id(name))
        self._ntf_interface = ntf_interface
        self._asynchronous = asynchronous

        # Create internal components.
        self._qdevice: QDevice = QDevice(self._node, qdevice_topology)
        self._local_ehi: EhiNodeInfo = LhiConverter.to_ehi(
            qdevice_topology, ntf_interface, latencies
        )

        host_latencies = HostLatencies(
            latencies.host_instr_time,
            latencies.host_peer_latency,
        )
        qnos_latencies = QnosLatencies(
            latencies.qnos_instr_time,
        )
        # TODO: decide if still needed
        netstack_latencies = NetstackLatencies(0)

        self._host = Host(
            self.host_comp, self._local_env, host_latencies, self._asynchronous
        )
        self._memmgr = MemoryManager(self.node.name, self._qdevice, self._local_ehi)
        self._qnos = Qnos(
            self.qnos_comp,
            self._local_env,
            self._memmgr,
            self._qdevice,
            qnos_latencies,
            self._ntf_interface,
            self._asynchronous,
        )
        self._netstack = Netstack(
            self.netstack_comp,
            self._local_env,
            self._memmgr,
            self._qdevice,
            netstack_latencies,
        )

        if scheduler is None:
            self._scheduler = Scheduler(
                self._node.name,
                self._host,
                self._qnos,
                self._netstack,
                self._memmgr,
                self._local_env,
                self._local_ehi,
                self._network_ehi,
            )
        else:
            self._scheduler = scheduler

    @property
    def node(self) -> ProcNodeComponent:
        return self._node

    @property
    def host_comp(self) -> HostComponent:
        return self.node.host_comp

    @property
    def qnos_comp(self) -> QnosComponent:
        return self.node.qnos_comp

    @property
    def netstack_comp(self) -> NetstackComponent:
        return self.node.netstack_comp

    @property
    def qdevice(self) -> QDevice:
        return self._qdevice

    @qdevice.setter
    def qdevice(self, qdevice) -> None:
        self._qdevice = qdevice
        self.qnos.qdevice = qdevice
        self.netstack.qdevice = qdevice

    @property
    def host(self) -> Host:
        return self._host

    @host.setter
    def host(self, host: Host) -> None:
        self._host = host

    @property
    def qnos(self) -> Qnos:
        return self._qnos

    @qnos.setter
    def qnos(self, qnos: Qnos) -> None:
        self._qnos = qnos

    @property
    def netstack(self) -> Netstack:
        return self._netstack

    @netstack.setter
    def netstack(self, netstack: Netstack) -> None:
        self._netstack = netstack

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    @property
    def scheduler(self) -> Scheduler:
        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler: Scheduler) -> None:
        self._scheduler = scheduler

    @property
    def local_ehi(self) -> EhiNodeInfo:
        return self._local_ehi

    @local_ehi.setter
    def local_ehi(self, local_ehi: EhiNodeInfo) -> None:
        self._local_ehi = local_ehi

    @property
    def network_ehi(self) -> EhiNetworkInfo:
        return self._network_ehi

    @network_ehi.setter
    def network_ehi(self, network_ehi: EhiNetworkInfo) -> None:
        self._network_ehi = network_ehi

    def connect_to(self, other: ProcNode) -> None:
        """Create connections between ports of this ProcNode and those of
        another ProcNode."""
        here = self.node.name
        there = other.node.name
        self.node.host_peer_out_port(there).connect(other.node.host_peer_in_port(here))
        self.node.host_peer_in_port(there).connect(other.node.host_peer_out_port(here))
        self.node.netstack_peer_out_port(there).connect(
            other.node.netstack_peer_in_port(here)
        )
        self.node.netstack_peer_in_port(there).connect(
            other.node.netstack_peer_out_port(here)
        )

    def start(self) -> None:
        assert self._host is not None
        assert self._qnos is not None
        assert self._netstack is not None
        super().start()
        self._host.start()
        self._qnos.start()
        self._netstack.start()
        self._scheduler.start()

    def stop(self) -> None:
        assert self._host is not None
        assert self._qnos is not None
        assert self._netstack is not None
        self._scheduler.stop()
        self._netstack.stop()
        self._qnos.stop()
        self._host.stop()
        super().stop()

    def submit_batch(self, batch_info: BatchInfo) -> None:
        self.scheduler.submit_batch(batch_info)

    def initialize_processes(self) -> None:
        self.scheduler.create_processes_for_batches()

    def add_process(self, process: QoalaProcess) -> None:
        self.memmgr.add_process(process)

    def get_batches(self) -> Dict[int, ProgramBatch]:
        return self.scheduler.get_batches()



from __future__ import annotations

from typing import Generator

import netsquid as ns
from netsquid.nodes import Node

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment, NetworkInfo
from qoala.runtime.message import Message
from qoala.sim.host.hostcomp import HostComponent
from qoala.sim.host.hostinterface import HostInterface


def create_hostcomp(num_other_nodes: int) -> HostComponent:
    node = Node(name="alice", ID=0)

    nodes = {id: f"node_{id}" for id in range(1, num_other_nodes + 1)}
    nodes[0] = "alice"
    env = NetworkInfo.with_nodes(nodes)

    return HostComponent(node, env)


def test_no_other_nodes():
    comp = create_hostcomp(num_other_nodes=0)

    # should have qnos_{in|out} and ntsk_{in|out} ports
    assert len(comp.ports) == 4
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "nstk_in" in comp.ports
    assert "nstk_out" in comp.ports


def test_one_other_node():
    comp = create_hostcomp(num_other_nodes=1)

    # should have 2 qnos ports + 2 nstk ports + 2 peer ports
    assert len(comp.ports) == 6
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "nstk_in" in comp.ports
    assert "nstk_out" in comp.ports

    assert "peer_node_1_in" in comp.ports
    assert "peer_node_1_out" in comp.ports

    # Test properties
    assert comp.peer_in_port("node_1") == comp.ports["peer_node_1_in"]
    assert comp.peer_out_port("node_1") == comp.ports["peer_node_1_out"]


def test_many_other_nodes():
    comp = create_hostcomp(num_other_nodes=5)

    # should have 2 qnos ports + 2 nstk ports + 5 * 2 peer ports
    assert len(comp.ports) == 14
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "nstk_in" in comp.ports
    assert "nstk_out" in comp.ports

    for i in range(1, 6):
        assert f"peer_node_{i}_in" in comp.ports
        assert f"peer_node_{i}_out" in comp.ports
        # Test properties
        assert comp.peer_in_port(f"node_{i}") == comp.ports[f"peer_node_{i}_in"]
        assert comp.peer_out_port(f"node_{i}") == comp.ports[f"peer_node_{i}_out"]


def test_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    env = NetworkInfo.with_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = HostComponent(alice, env)
    bob_comp = HostComponent(bob, env)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))

    class AliceHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message("hello"))

    class BobHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            print(f"{self.name}: received msg with content: {msg.content}")

    alice_intf = AliceHostInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobHostInterface(bob_comp, LocalEnvironment(env, bob.ID))

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


def test_three_way_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    charlie = Node(name="charlie", ID=2)
    env = NetworkInfo.with_nodes(
        {alice.ID: alice.name, bob.ID: bob.name, charlie.ID: charlie.name}
    )

    alice_comp = HostComponent(alice, env)
    bob_comp = HostComponent(bob, env)
    charlie_comp = HostComponent(charlie, env)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))
    alice_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("alice"))
    bob_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("bob"))
    bob_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("bob"))

    class AliceHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message("hello bob"))
            self.send_peer_msg("charlie", Message("hello charlie"))

    class BobHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            print(f"{self.name}: received msg with content: {msg.content}")

    class CharlieHostInterface(HostInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            print(f"{self.name}: received msg with content: {msg.content}")

    alice_intf = AliceHostInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobHostInterface(bob_comp, LocalEnvironment(env, bob.ID))
    charlie_intf = CharlieHostInterface(charlie_comp, LocalEnvironment(env, charlie.ID))

    alice_intf.start()
    bob_intf.start()
    charlie_intf.start()

    ns.sim_run()


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()

    test_connection()
    test_three_way_connection()



from __future__ import annotations

from typing import Dict, List

from netsquid.components.instructions import (
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
)
from netsquid.components.models.qerrormodels import DepolarNoiseModel, T1T2NoiseModel

from qoala.lang.common import MultiQubit

# Ignore type since whole 'config' module is ignored by mypy
from qoala.runtime.config import NVQDeviceConfig  # type: ignore
from qoala.runtime.lhi import LhiGateInfo, LhiQubitInfo, LhiTopology


class LhiTopologyBuilderForOldNV:
    @classmethod
    def from_nv_config(cls, cfg: NVQDeviceConfig) -> LhiTopology:
        qubit_infos: Dict[int, LhiQubitInfo] = {}

        comm_info = LhiQubitInfo(
            is_communication=True,
            error_model=T1T2NoiseModel,
            error_model_kwargs={"T1": cfg.electron_T1, "T2": cfg.electron_T2},
        )
        mem_info = LhiQubitInfo(
            is_communication=False,
            error_model=T1T2NoiseModel,
            error_model_kwargs={"T1": cfg.carbon_T1, "T2": cfg.carbon_T2},
        )
        qubit_infos[0] = comm_info
        for i in range(1, cfg.num_qubits):
            qubit_infos[i] = mem_info

        single_gate_infos: Dict[int, List[LhiGateInfo]] = {}
        comm_gates: List[LhiGateInfo] = []
        comm_gates += [
            LhiGateInfo(
                instruction=INSTR_INIT,
                duration=cfg.electron_init,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": cfg.electron_init_depolar_prob},
            )
        ]
        comm_gates += [
            # TODO: how to implement 'hack for imperfect measurements'?
            # (see bulid_nv_qprocessor in build.py)
            LhiGateInfo(
                instruction=INSTR_MEASURE,
                duration=cfg.measure,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": 0},
            )
        ]
        comm_gates += [
            LhiGateInfo(
                instruction=instr,
                duration=dur,
                error_model=DepolarNoiseModel,
                error_model_kwargs={
                    "depolar_rate": cfg.electron_single_qubit_depolar_prob
                },
            )
            for instr, dur in zip(
                [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
                [cfg.electron_rot_x, cfg.electron_rot_y, cfg.electron_rot_z],
            )
        ]
        mem_gates: List[LhiGateInfo] = []
        mem_gates += [
            LhiGateInfo(
                instruction=INSTR_INIT,
                duration=cfg.carbon_init,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": cfg.carbon_init_depolar_prob},
            )
        ]
        mem_gates += [
            LhiGateInfo(
                instruction=instr,
                duration=dur,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": cfg.carbon_z_rot_depolar_prob},
            )
            for instr, dur in zip(
                [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
                [cfg.carbon_rot_x, cfg.carbon_rot_y, cfg.carbon_rot_z],
            )
        ]

        single_gate_infos[0] = comm_gates
        for i in range(1, cfg.num_qubits):
            single_gate_infos[i] = mem_gates

        multi_gate_infos: Dict[MultiQubit, List[LhiGateInfo]] = {}
        gate_infos: List[LhiGateInfo] = [
            LhiGateInfo(
                instruction=instr,
                duration=dur,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": cfg.ec_gate_depolar_prob},
            )
            for instr, dur in zip(
                [INSTR_CXDIR, INSTR_CYDIR],
                [cfg.ec_controlled_dir_x, cfg.ec_controlled_dir_y],
            )
        ]

        for i in range(1, cfg.num_qubits):
            multi = MultiQubit([0, i])
            multi_gate_infos[multi] = gate_infos

        return LhiTopology(
            qubit_infos=qubit_infos,
            single_gate_infos=single_gate_infos,
            multi_gate_infos=multi_gate_infos,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from qoala.lang.hostlang import BasicBlockType
from qoala.runtime.task import BlockTask


@dataclass
class TaskScheduleEntry:
    task: BlockTask
    timestamp: Optional[float] = None
    prev: Optional[BlockTask] = None

    def is_cpu_task(self) -> bool:
        return self.task.typ == BasicBlockType.CL or self.task.typ == BasicBlockType.CC

    def is_qpu_task(self) -> bool:
        return self.task.typ == BasicBlockType.QL or self.task.typ == BasicBlockType.QC


@dataclass
class LinkSlotInfo:
    offset1: float  # time for first pair
    offset2: float  # time for second pair
    period: float  # time until next cycle (with again two pairs)


@dataclass
class QcSlotInfo:
    links: Dict[int, LinkSlotInfo]


@dataclass
class TaskSchedule:
    entries: List[TaskScheduleEntry]

    @classmethod
    def _compute_timestamps(
        cls, task_list: List[BlockTask], qc_slot_info: Optional[QcSlotInfo]
    ) -> List[float]:
        # Get QC task indices
        qc_indices: List[int] = []
        for i, task in enumerate(task_list):
            if task.typ == BasicBlockType.QC:
                qc_indices.append(i)

        # list of timestamps for each task (same order as tasks in task_list)
        timestamps: List[float] = []
        # Naive approach: first schedule all tasks consecutively. Then, move the first
        # QC task forward by X until it aligns with a qc_slot. Also move all tasks
        # coming fater this QC task by X (so they are still consecutive).
        # Repeat for the next QC task in the list.
        time = 0.0
        for i, task in enumerate(task_list):
            timestamps.append(time)
            if task.duration:
                time += task.duration

        if qc_slot_info is not None:
            # remote ID -> (slot, offsetX)
            curr_slots: Dict[int, Tuple[float, int]] = {
                remote_id: (qc_slot_info.links[remote_id].offset1, 1)
                for remote_id in qc_slot_info.links.keys()
            }

            for index in qc_indices:
                remote_id = task_list[index].remote_id
                assert remote_id is not None
                info = qc_slot_info.links[remote_id]
                curr_slot, offset_index = curr_slots[remote_id]
                while curr_slot <= timestamps[index]:
                    if offset_index == 1:
                        offset_index = 2
                        curr_slot += info.offset2 - info.offset1
                    else:
                        offset_index = 1
                        curr_slot += info.period - (info.offset2 - info.offset1)
                curr_slots[remote_id] = curr_slot, offset_index

                delta = curr_slot - timestamps[index]
                for i in range(index, len(timestamps)):
                    if timestamps[i] is not None:
                        timestamps[i] += delta
        return timestamps

    @classmethod
    def consecutive(
        cls, task_list: List[BlockTask], qc_slot_info: Optional[QcSlotInfo] = None
    ) -> TaskSchedule:
        entries: List[TaskScheduleEntry] = []

        if qc_slot_info is not None:
            timestamps = cls._compute_timestamps(task_list, qc_slot_info)

        for i, task in enumerate(task_list):
            if qc_slot_info is not None and task.typ == BasicBlockType.QC:
                time = timestamps[i]
            else:
                time = None
            entry = TaskScheduleEntry(task=task, timestamp=time, prev=None)
            entries.append(entry)

        for i in range(len(entries) - 1):
            e1 = entries[i]
            e2 = entries[i + 1]
            if e1.is_cpu_task() != e2.is_cpu_task():
                e2.prev = e1.task

        return TaskSchedule(entries)

    @classmethod
    def consecutive_timestamps(
        cls,
        task_list: List[BlockTask],
        qc_slot_info: Optional[QcSlotInfo] = None,
    ) -> TaskSchedule:
        entries: List[TaskScheduleEntry] = []

        timestamps = cls._compute_timestamps(task_list, qc_slot_info)

        for i, task in enumerate(task_list):
            time = timestamps[i]
            entries.append(TaskScheduleEntry(task, time, None))

        return TaskSchedule(entries)

    @property
    def cpu_schedule(self) -> TaskSchedule:
        entries = [e for e in self.entries if e.is_cpu_task()]
        return TaskSchedule(entries)

    @property
    def qpu_schedule(self) -> TaskSchedule:
        entries = [e for e in self.entries if e.is_qpu_task()]
        return TaskSchedule(entries)

    def __str__(self) -> str:
        return ScheduleWriter(self).write()


class ScheduleWriter:
    def __init__(self, schedule: TaskSchedule) -> None:
        self._timeline = "time "
        self._cpu_task_str = "CPU  "
        self._qpu_task_str = "QPU  "
        self._cpu_entries = schedule.cpu_schedule.entries
        self._qpu_entries = schedule.qpu_schedule.entries
        self._entry_width = 12

    def _entry_content(self, task: BlockTask) -> str:
        if task.duration:
            return f"{task.block_name} ({task.typ.name}, {int(task.duration)})"
        else:
            return f"{task.block_name} ({task.typ.name})"

    def _add_cpu_entry(self, cpu_time: Optional[float], cpu_task: BlockTask) -> None:
        width = max(len(cpu_task.block_name), len(str(cpu_time))) + self._entry_width
        if cpu_time is None:
            cpu_time = "<none>"  # type: ignore
        self._timeline += f"{cpu_time:<{width}}"
        entry = self._entry_content(cpu_task)
        self._cpu_task_str += f"{entry:<{width}}"
        self._qpu_task_str += " " * width

    def _add_qpu_entry(self, qpu_time: Optional[float], qpu_task: BlockTask) -> None:
        width = max(len(qpu_task.block_name), len(str(qpu_time))) + self._entry_width
        if qpu_time is None:
            qpu_time = "<none>"  # type: ignore
        self._timeline += f"{qpu_time:<{width}}"
        entry = self._entry_content(qpu_task)
        self._cpu_task_str += " " * width
        self._qpu_task_str += f"{entry:<{width}}"

    def _add_double_entry(
        self, time: Optional[float], cpu_task: BlockTask, qpu_task: BlockTask
    ) -> None:
        width = (
            max(len(cpu_task.block_name), len(qpu_task.block_name), len(str(time)))
            + self._entry_width
        )
        self._timeline += f"{time:<{width}}"
        cpu_entry = self._entry_content(cpu_task)
        qpu_entry = self._entry_content(qpu_task)
        self._cpu_task_str += f"{cpu_entry:<{width}}"
        self._qpu_task_str += f"{qpu_entry:<{width}}"

    def write(self) -> str:
        cpu_index = 0
        qpu_index = 0
        cpu_done = False
        qpu_done = False
        while True:
            try:
                cpu_entry = self._cpu_entries[cpu_index]
                cpu_task = cpu_entry.task
                cpu_time = cpu_entry.timestamp
            except IndexError:
                cpu_done = True
            try:
                qpu_entry = self._qpu_entries[qpu_index]
                qpu_task = qpu_entry.task
                qpu_time = qpu_entry.timestamp
            except IndexError:
                qpu_done = True
            if cpu_done and qpu_done:
                break
            if qpu_done:  # cpu_done is False
                cpu_index += 1
                self._add_cpu_entry(cpu_time, cpu_task)
            elif cpu_done:  # qpu_done is False
                qpu_index += 1
                self._add_qpu_entry(qpu_time, qpu_task)
            else:  # both not done
                if qpu_time is None:
                    cpu_index += 1
                    self._add_cpu_entry(cpu_time, cpu_task)
                elif cpu_time is None:
                    qpu_index += 1
                    self._add_qpu_entry(qpu_time, qpu_task)
                elif cpu_time < qpu_time:
                    cpu_index += 1
                    self._add_cpu_entry(cpu_time, cpu_task)
                elif qpu_time < cpu_time:
                    qpu_index += 1
                    self._add_qpu_entry(qpu_time, qpu_task)
                else:  # times equal
                    cpu_index += 1
                    qpu_index += 1
                    self._add_double_entry(cpu_time, cpu_task, qpu_task)
        return self._timeline + "\n" + self._cpu_task_str + "\n" + self._qpu_task_str



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import List, Optional

import netsquid as ns
import pytest

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    LinkBetweenNodesConfig,
    LinkConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.program import BatchInfo, BatchResult, ProgramInput
from qoala.runtime.schedule import TaskSchedule
from qoala.sim.build import build_network
from qoala.util.math import fidelity_to_prob_max_mixed


def create_network_info(names: List[str]) -> NetworkInfo:
    env = NetworkInfo.with_nodes({i: name for i, name in enumerate(names)})
    env.set_global_schedule([0, 1, 2])
    env.set_timeslot_len(1e6)
    return env


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def create_batch(
    program: QoalaProgram,
    unit_module: UnitModule,
    inputs: List[ProgramInput],
    num_iterations: int,
) -> BatchInfo:
    return BatchInfo(
        program=program,
        unit_module=unit_module,
        inputs=inputs,
        num_iterations=num_iterations,
        deadline=0,
    )


@dataclass
class QkdResult:
    alice_result: BatchResult
    bob_result: BatchResult


def run_qkd(
    num_iterations: int,
    link_fidelity: float,
    alice_file: str,
    bob_file: str,
    num_pairs: Optional[int] = None,
):
    ns.sim_reset()

    num_qubits = 3
    network_info = create_network_info(names=["alice", "bob"])
    alice_id = network_info.get_node_id("alice")
    bob_id = network_info.get_node_id("bob")

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits)

    link_cfg = LinkConfig.depolarise_config(fidelity=link_fidelity, state_delay=1000)
    link_between_cfg = LinkBetweenNodesConfig(
        node_id1=alice_id, node_id2=bob_id, link_config=link_cfg
    )
    network_cfg = ProcNodeNetworkConfig(
        nodes=[alice_node_cfg, bob_node_cfg], links=[link_between_cfg]
    )
    network = build_network(network_cfg, network_info)
    alice_procnode = network.nodes["alice"]
    bob_procnode = network.nodes["bob"]

    alice_program = load_program(alice_file)
    if num_pairs is not None:
        alice_inputs = [
            ProgramInput({"bob_id": bob_id, "num_pairs": num_pairs})
            for _ in range(num_iterations)
        ]
    else:
        alice_inputs = [ProgramInput({"bob_id": bob_id}) for _ in range(num_iterations)]

    alice_unit_module = UnitModule.from_full_ehi(alice_procnode.memmgr.get_ehi())
    alice_batch = create_batch(
        alice_program, alice_unit_module, alice_inputs, num_iterations
    )
    alice_procnode.submit_batch(alice_batch)
    alice_procnode.initialize_processes()
    alice_tasks = alice_procnode.scheduler.get_tasks_to_schedule()
    alice_schedule = TaskSchedule.consecutive(alice_tasks)
    alice_procnode.scheduler.upload_schedule(alice_schedule)

    bob_program = load_program(bob_file)
    if num_pairs is not None:
        bob_inputs = [
            ProgramInput({"alice_id": alice_id, "num_pairs": num_pairs})
            for _ in range(num_iterations)
        ]
    else:
        bob_inputs = [
            ProgramInput({"alice_id": alice_id}) for _ in range(num_iterations)
        ]

    bob_unit_module = UnitModule.from_full_ehi(bob_procnode.memmgr.get_ehi())
    bob_batch = create_batch(bob_program, bob_unit_module, bob_inputs, num_iterations)
    bob_procnode.submit_batch(bob_batch)
    bob_procnode.initialize_processes()
    bob_tasks = bob_procnode.scheduler.get_tasks_to_schedule()
    bob_schedule = TaskSchedule.consecutive(bob_tasks)
    bob_procnode.scheduler.upload_schedule(bob_schedule)

    network.start()
    ns.sim_run()

    # only one batch (ID = 0), so get value at index 0
    alice_result = alice_procnode.scheduler.get_batch_results()[0]
    bob_result = bob_procnode.scheduler.get_batch_results()[0]

    return QkdResult(alice_result, bob_result)


def test_qkd_md_1pair():
    ns.sim_reset()

    num_iterations = 1000

    link_fidelity = 0.7

    alice_file = "qkd_md_1pair_alice.iqoala"
    bob_file = "qkd_md_1pair_bob.iqoala"

    qkd_result = run_qkd(num_iterations, link_fidelity, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    count_equal_outcomes = 0
    for alice, bob in zip(alice_outcomes, bob_outcomes):
        if alice["m0"] == bob["m0"]:
            count_equal_outcomes += 1

    # We used a link fidelity of 0.7 for a depolarising sampler.
    # This results in a produced state of 0.4 * <maximally mixed> + 0.6 * Phi+.
    assert fidelity_to_prob_max_mixed(0.7) == pytest.approx(0.4)
    # Hence we expect the ratio of pairs with equal outcomes to be
    # 0.5 * 0.4                     +    1.0 * 0.6                   = 0.8
    # (mixed state -> 50% success)       (Phi+ -> 100% success)
    assert (count_equal_outcomes / num_iterations) <= 0.85
    assert (count_equal_outcomes / num_iterations) >= 0.75


if __name__ == "__main__":
    test_qkd_md_1pair()



from qoala.lang.ehi import EhiBuilder, UnitModule
from qoala.lang.parse import LocalRoutineParser
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.lang.routine import LocalRoutine
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.sim.process import QoalaProcess


def create_local_routine() -> LocalRoutine:
    text = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 
    keeps:
    request: 
  NETQASM_START
    set Q0 {my_value}
  NETQASM_END
    """

    parsed = LocalRoutineParser(text).parse()
    return parsed["subrt1"]


def create_process(program: QoalaProgram) -> QoalaProcess:
    ehi = EhiBuilder.perfect_uniform(1, None, [], 0, [], 0)
    unit_module = UnitModule.from_full_ehi(ehi)

    instance = ProgramInstance(
        pid=0,
        program=program,
        inputs=ProgramInput({}),
        unit_module=unit_module,
        block_tasks=[],
    )
    mem = ProgramMemory(pid=0)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def test1():
    routine = create_local_routine()

    program = QoalaProgram(
        blocks=[], local_routines={"subrt1": routine}, meta=ProgramMeta.empty("")
    )
    process = create_process(program)

    assert len(process.get_all_local_routines()) == 1
    assert process.get_local_routine("subrt1") == routine
    assert len(process.qnos_mem.get_all_running_local_routines()) == 0


if __name__ == "__main__":
    test1()



from __future__ import annotations

from netsquid.components.component import Component, Port
from netsquid.nodes import Node


class QnosComponent(Component):
    """NetSquid component representing a QNodeOS instance.

    Subcomponent of a ProcNodeComponent.

    This is a static container for QNodeOS-related components and ports.
    Behavior of a QNodeOS instance is modeled in the `Qnos` class,
    which is a subclass of `Protocol`.
    """

    def __init__(self, node: Node) -> None:
        super().__init__(name=f"{node.name}_qnos")
        self._node = node

        # Ports for communicating with Host
        self.add_ports(["host_out", "host_in"])
        self.add_ports(["nstk_out", "nstk_in"])

        # Separate channel for "memory freed" signals
        self.add_ports(["nstk_mem_out", "nstk_mem_in"])

    @property
    def host_in_port(self) -> Port:
        return self.ports["host_in"]

    @property
    def host_out_port(self) -> Port:
        return self.ports["host_out"]

    @property
    def netstack_in_port(self) -> Port:
        return self.ports["nstk_in"]

    @property
    def netstack_out_port(self) -> Port:
        return self.ports["nstk_out"]

    @property
    def netstack_mem_in_port(self) -> Port:
        return self.ports["nstk_mem_in"]

    @property
    def netstack_mem_out_port(self) -> Port:
        return self.ports["nstk_mem_out"]

    @property
    def node(self) -> Node:
        return self._node



import os

from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.models.qerrormodels import DepolarNoiseModel, T1T2NoiseModel
from netsquid_magic.state_delivery_sampler import (
    DepolariseWithFailureStateSamplerFactory,
    PerfectStateSamplerFactory,
)

from qoala.lang.common import MultiQubit
from qoala.runtime.config import (
    LatenciesConfig,
    LinkConfig,
    NVQDeviceConfig,
    TopologyConfig,
)
from qoala.runtime.lhi import (
    LhiGateInfo,
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiQubitInfo,
    LhiTopology,
    LhiTopologyBuilder,
)
from qoala.runtime.lhi_nv_compat import LhiTopologyBuilderForOldNV


def relative_path(path: str) -> str:
    return os.path.join(os.getcwd(), os.path.dirname(__file__), path)


def test_topology():
    comm_qubit_info = LhiQubitInfo(
        is_communication=True,
        error_model=T1T2NoiseModel,
        error_model_kwargs={"T1": 1e3, "T2": 2e3},
    )
    mem_qubit_info = LhiQubitInfo(
        is_communication=False,
        error_model=T1T2NoiseModel,
        error_model_kwargs={"T1": 1e3, "T2": 2e3},
    )
    gate_x_info = LhiGateInfo(
        instruction=INSTR_X,
        duration=1e4,
        error_model=DepolarNoiseModel,
        error_model_kwargs={
            "depolar_rate": 0.2,
            "time_independent": True,
        },
    )
    gate_y_info = LhiGateInfo(
        instruction=INSTR_Y,
        duration=1e4,
        error_model=DepolarNoiseModel,
        error_model_kwargs={
            "depolar_rate": 0.2,
            "time_independent": True,
        },
    )
    cnot_gate_info = LhiGateInfo(
        instruction=INSTR_CNOT,
        duration=1e4,
        error_model=DepolarNoiseModel,
        error_model_kwargs={
            "depolar_rate": 0.2,
            "time_independent": True,
        },
    )

    qubit_infos = {0: comm_qubit_info, 1: mem_qubit_info}
    single_gate_infos = {0: [gate_x_info, gate_y_info], 1: [gate_x_info]}
    multi_gate_infos = {MultiQubit([0, 1]): [cnot_gate_info]}
    topology = LhiTopology(
        qubit_infos=qubit_infos,
        single_gate_infos=single_gate_infos,
        multi_gate_infos=multi_gate_infos,
    )

    assert topology.qubit_infos[0].is_communication
    assert not topology.qubit_infos[1].is_communication

    q0_gates = [info.instruction for info in topology.single_gate_infos[0]]
    assert q0_gates == [INSTR_X, INSTR_Y]

    q1_gates = [info.instruction for info in topology.single_gate_infos[1]]
    assert q1_gates == [INSTR_X]

    q01_gates = [
        info.instruction for info in topology.multi_gate_infos[MultiQubit([0, 1])]
    ]
    assert q01_gates == [INSTR_CNOT]


def test_topology_from_config():
    cfg = TopologyConfig.from_file(relative_path("configs/topology_cfg_4.yaml"))
    topology = LhiTopologyBuilder.from_config(cfg)

    assert topology.qubit_infos[0].is_communication
    assert not topology.qubit_infos[1].is_communication

    q0_gates = [info.instruction for info in topology.single_gate_infos[0]]
    assert q0_gates == [INSTR_X, INSTR_Y]

    q1_gates = [info.instruction for info in topology.single_gate_infos[1]]
    assert q1_gates == [INSTR_X]

    q01_gates = [
        info.instruction for info in topology.multi_gate_infos[MultiQubit([0, 1])]
    ]
    assert q01_gates == [INSTR_CNOT]


def test_topology_from_config_2():
    cfg = TopologyConfig.from_file(relative_path("configs/topology_cfg_6.yaml"))
    topology = LhiTopologyBuilder.from_config(cfg)

    assert topology.qubit_infos[0].is_communication
    for i in [1, 2, 3]:
        assert not topology.qubit_infos[i].is_communication

    q0_gates = [info.instruction for info in topology.single_gate_infos[0]]
    assert q0_gates == [INSTR_X, INSTR_Y, INSTR_Z]

    for i in [1, 2, 3]:
        qi_gates = [info.instruction for info in topology.single_gate_infos[i]]
        assert qi_gates == [INSTR_X, INSTR_Y]

    for i in [1, 2, 3]:
        q0i_gates = [
            info.instruction for info in topology.multi_gate_infos[MultiQubit([0, i])]
        ]
        assert q0i_gates == [INSTR_CNOT]


def test_topology_from_nv_config():
    cfg = NVQDeviceConfig.perfect_config(num_qubits=2)
    topology = LhiTopologyBuilderForOldNV.from_nv_config(cfg)

    assert topology.qubit_infos[0].is_communication
    assert not topology.qubit_infos[1].is_communication

    q0_gates = [info.instruction for info in topology.single_gate_infos[0]]
    assert all(
        instr in q0_gates
        for instr in [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z, INSTR_INIT, INSTR_MEASURE]
    )

    q1_gates = [info.instruction for info in topology.single_gate_infos[1]]
    assert all(
        instr in q1_gates
        for instr in [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z, INSTR_INIT]
    )

    q01_gates = [
        info.instruction for info in topology.multi_gate_infos[MultiQubit([0, 1])]
    ]
    assert all(instr in q01_gates for instr in [INSTR_CXDIR, INSTR_CYDIR])

    assert topology.find_single_gate(0, INSTR_ROT_X).duration == cfg.electron_rot_x
    assert topology.find_single_gate(0, INSTR_ROT_X).error_model_kwargs == {
        "depolar_rate": 0
    }


def test_find_gates():
    cfg = TopologyConfig.from_file(relative_path("configs/topology_cfg_6.yaml"))
    topology = LhiTopologyBuilder.from_config(cfg)

    for i in range(4):
        assert topology.find_single_gate(i, INSTR_X) is not None
        assert topology.find_single_gate(i, INSTR_Y) is not None
        if i == 0:
            assert topology.find_single_gate(i, INSTR_Z) is not None
        else:
            assert topology.find_single_gate(i, INSTR_Z) is None
    assert topology.find_single_gate(4, INSTR_X) is None

    for i in range(1, 4):
        assert topology.find_multi_gate([0, i], INSTR_CNOT) is not None
        assert topology.find_multi_gate([i, 0], INSTR_CNOT) is None

    assert topology.find_multi_gate([0, 0], INSTR_CNOT) is None


def test_perfect_qubit():
    qubit_info = LhiTopologyBuilder.perfect_qubit(is_communication=True)
    assert qubit_info.is_communication
    assert qubit_info.error_model == T1T2NoiseModel
    assert qubit_info.error_model_kwargs == {"T1": 0, "T2": 0}


def test_t1t2_qubit():
    qubit_info = LhiTopologyBuilder.t1t2_qubit(is_communication=True, t1=1e3, t2=1e4)
    assert qubit_info.is_communication
    assert qubit_info.error_model == T1T2NoiseModel
    assert qubit_info.error_model_kwargs == {"T1": 1e3, "T2": 1e4}


def test_perfect_gates():
    duration = 5e3
    instructions = [INSTR_X, INSTR_Y, INSTR_Z]
    gate_infos = LhiTopologyBuilder.perfect_gates(duration, instructions)

    assert gate_infos == [
        LhiGateInfo(
            instruction=instr,
            duration=duration,
            error_model=DepolarNoiseModel,
            error_model_kwargs={"depolar_rate": 0},
        )
        for instr in instructions
    ]


def test_depolar_gates():
    duration = 5e3
    instructions = [INSTR_X, INSTR_Y, INSTR_Z]
    gate_infos = LhiTopologyBuilder.depolar_gates(
        duration, instructions, depolar_rate=0.4
    )

    assert gate_infos == [
        LhiGateInfo(
            instruction=instr,
            duration=duration,
            error_model=DepolarNoiseModel,
            error_model_kwargs={"depolar_rate": 0.4},
        )
        for instr in instructions
    ]


def test_perfect_uniform():
    num_qubits = 3
    single_qubit_instructions = [INSTR_X, INSTR_Y, INSTR_Z]
    single_gate_duration = 5e3
    two_qubit_instructions = [INSTR_CNOT, INSTR_CZ]
    two_gate_duration = 2e5
    topology = LhiTopologyBuilder.perfect_uniform(
        num_qubits,
        single_qubit_instructions,
        single_gate_duration,
        two_qubit_instructions,
        two_gate_duration,
    )

    assert topology.qubit_infos == {
        i: LhiQubitInfo(
            is_communication=True,
            error_model=T1T2NoiseModel,
            error_model_kwargs={"T1": 0, "T2": 0},
        )
        for i in range(num_qubits)
    }

    assert topology.single_gate_infos == {
        i: [
            LhiGateInfo(
                instruction=instr,
                duration=single_gate_duration,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": 0},
            )
            for instr in single_qubit_instructions
        ]
        for i in range(num_qubits)
    }

    assert topology.multi_gate_infos == {
        MultiQubit([i, j]): [
            LhiGateInfo(
                instruction=instr,
                duration=two_gate_duration,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": 0},
            )
            for instr in two_qubit_instructions
        ]
        for i in range(num_qubits)
        for j in range(num_qubits)
        if i != j
    }


def test_build_fully_uniform():
    qubit_info = LhiQubitInfo(
        is_communication=True,
        error_model=T1T2NoiseModel,
        error_model_kwargs={"T1": 1e6, "T2": 2e6},
    )
    single_gate_infos = [
        LhiGateInfo(
            instruction=instr,
            duration=5e3,
            error_model=DepolarNoiseModel,
            error_model_kwargs={
                "depolar_rate": 0.2,
                "time_independent": True,
            },
        )
        for instr in [INSTR_X, INSTR_Y, INSTR_Z]
    ]
    two_gate_infos = [
        LhiGateInfo(
            instruction=INSTR_CNOT,
            duration=2e4,
            error_model=DepolarNoiseModel,
            error_model_kwargs={
                "depolar_rate": 0.2,
                "time_independent": True,
            },
        )
    ]
    topology = LhiTopologyBuilder.fully_uniform(
        num_qubits=3,
        qubit_info=qubit_info,
        single_gate_infos=single_gate_infos,
        two_gate_infos=two_gate_infos,
    )

    assert len(topology.qubit_infos) == 3
    for i in range(3):
        assert topology.qubit_infos[i].is_communication
        single_gates = [info.instruction for info in topology.single_gate_infos[i]]
        assert INSTR_X in single_gates
        assert INSTR_Y in single_gates
        assert INSTR_Z in single_gates

    for i in range(3):
        for j in range(3):
            if i == j:
                continue
            multi = MultiQubit([i, j])
            gates = [info.instruction for info in topology.multi_gate_infos[multi]]
            assert INSTR_CNOT in gates


def test_perfect_star():
    num_qubits = 3
    comm_instructions = [INSTR_X, INSTR_Y, INSTR_Z]
    comm_duration = 5e3
    mem_instructions = [INSTR_X, INSTR_Y]
    mem_duration = 1e4
    two_instructions = [INSTR_CNOT, INSTR_CZ]
    two_duration = 2e5
    topology = LhiTopologyBuilder.perfect_star(
        num_qubits,
        comm_instructions,
        comm_duration,
        mem_instructions,
        mem_duration,
        two_instructions,
        two_duration,
    )

    assert topology.qubit_infos[0] == LhiTopologyBuilder.perfect_qubit(
        is_communication=True
    )

    for i in range(1, num_qubits):
        assert topology.qubit_infos[i] == LhiTopologyBuilder.perfect_qubit(
            is_communication=False
        )

    assert topology.single_gate_infos[0] == LhiTopologyBuilder.perfect_gates(
        comm_duration, [INSTR_X, INSTR_Y, INSTR_Z]
    )
    for i in range(1, num_qubits):
        assert topology.single_gate_infos[i] == LhiTopologyBuilder.perfect_gates(
            mem_duration, [INSTR_X, INSTR_Y]
        )

    for i in range(1, num_qubits):
        assert topology.multi_gate_infos[
            MultiQubit([0, i])
        ] == LhiTopologyBuilder.perfect_gates(two_duration, [INSTR_CNOT, INSTR_CZ])


def test_generic_t1t2_star():
    num_qubits = 3
    comm_t1 = 1e8
    comm_t2 = 2e8
    mem_t1 = 1e9
    mem_t2 = 2e9
    comm_instructions = [INSTR_X, INSTR_Y, INSTR_Z]
    comm_duration = 5e3
    comm_instr_depolar_rate = 0.2
    mem_instructions = [INSTR_X, INSTR_Y]
    mem_duration = 1e4
    mem_instr_depolar_rate = 0.3
    two_instructions = [INSTR_CNOT, INSTR_CZ]
    two_duration = 2e5
    two_instr_depolar_rate = 0.4
    topology = LhiTopologyBuilder.generic_t1t2_star(
        num_qubits,
        comm_t1,
        comm_t2,
        mem_t1,
        mem_t2,
        comm_instructions,
        comm_duration,
        comm_instr_depolar_rate,
        mem_instructions,
        mem_duration,
        mem_instr_depolar_rate,
        two_instructions,
        two_duration,
        two_instr_depolar_rate,
    )

    assert topology.qubit_infos[0] == LhiTopologyBuilder.t1t2_qubit(
        is_communication=True, t1=comm_t1, t2=comm_t2
    )

    for i in range(1, num_qubits):
        assert topology.qubit_infos[i] == LhiTopologyBuilder.t1t2_qubit(
            is_communication=False, t1=mem_t1, t2=mem_t2
        )

    assert topology.single_gate_infos[0] == LhiTopologyBuilder.depolar_gates(
        comm_duration, [INSTR_X, INSTR_Y, INSTR_Z], comm_instr_depolar_rate
    )
    for i in range(1, num_qubits):
        assert topology.single_gate_infos[i] == LhiTopologyBuilder.depolar_gates(
            mem_duration, [INSTR_X, INSTR_Y], mem_instr_depolar_rate
        )

    for i in range(1, num_qubits):
        assert topology.multi_gate_infos[
            MultiQubit([0, i])
        ] == LhiTopologyBuilder.depolar_gates(
            two_duration, [INSTR_CNOT, INSTR_CZ], two_instr_depolar_rate
        )


def test_latencies_from_config():
    cfg = LatenciesConfig(
        host_instr_time=2,
        qnos_instr_time=3,
        host_peer_latency=4,
    )
    latencies = LhiLatencies.from_config(cfg)

    assert latencies.host_instr_time == 2
    assert latencies.qnos_instr_time == 3
    assert latencies.host_peer_latency == 4


def test_link_from_config():
    cfg = LinkConfig.from_file(relative_path("configs/link_cfg_1.yaml"))
    link = LhiLinkInfo.from_config(cfg)

    assert link.state_delay == 750
    assert link.sampler_factory == PerfectStateSamplerFactory
    assert link.sampler_kwargs == {"cycle_time": 25}


def test_link_from_config_2():
    cfg = LinkConfig.from_file(relative_path("configs/link_cfg_2.yaml"))
    link = LhiLinkInfo.from_config(cfg)

    assert link.state_delay == 750
    assert link.sampler_factory == DepolariseWithFailureStateSamplerFactory
    assert link.sampler_kwargs == {
        "cycle_time": 10,
        "prob_max_mixed": 0.3,
        "prob_success": 0.1,
    }


def test_network_lhi():
    network_lhi = LhiNetworkInfo.perfect_fully_connected(
        node_ids=[0, 1, 2], duration=1000
    )

    assert network_lhi.get_link(0, 1) == LhiLinkInfo.perfect(duration=1000)
    assert network_lhi.get_link(0, 2) == LhiLinkInfo.perfect(duration=1000)
    assert network_lhi.get_link(1, 2) == LhiLinkInfo.perfect(duration=1000)


if __name__ == "__main__":
    test_topology()
    test_topology_from_config()
    test_topology_from_config_2()
    test_topology_from_nv_config()
    test_find_gates()
    test_perfect_qubit()
    test_t1t2_qubit()
    test_perfect_gates()
    test_depolar_gates()
    test_perfect_uniform()
    test_build_fully_uniform()
    test_perfect_star()
    test_generic_t1t2_star()
    test_latencies_from_config()
    test_link_from_config()
    test_link_from_config_2()
    test_network_lhi()



from __future__ import annotations

import os
from typing import Any, Dict, Type

from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_H,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.instructions import Instruction as NetSquidInstruction
from netsquid.components.models.qerrormodels import (
    DepolarNoiseModel,
    QuantumErrorModel,
    T1T2NoiseModel,
)
from netsquid_magic.state_delivery_sampler import (
    DepolariseWithFailureStateSamplerFactory,
    PerfectStateSamplerFactory,
)

from qoala.lang.common import MultiQubit
from qoala.runtime.config import (
    BaseModel,
    DepolariseSamplerConfig,
    GateConfig,
    GateConfigRegistry,
    GateDepolariseConfig,
    GateNoiseConfigInterface,
    InstrConfigRegistry,
    LatenciesConfig,
    LinkBetweenNodesConfig,
    LinkConfig,
    MultiGateConfig,
    PerfectSamplerConfig,
    ProcNodeConfig,
    QubitConfig,
    QubitConfigRegistry,
    QubitIdConfig,
    QubitNoiseConfigInterface,
    QubitT1T2Config,
    SingleGateConfig,
    TopologyConfig,
)
from qoala.util.math import fidelity_to_prob_max_mixed


def relative_path(path: str) -> str:
    return os.path.join(os.getcwd(), os.path.dirname(__file__), path)


def test_qubit_t1t2_config():
    cfg = QubitT1T2Config(T1=1e6, T2=3e6)

    assert cfg.T1 == 1e6
    assert cfg.T2 == 3e6


def test_qubit_t1t2_config_file():
    cfg = QubitT1T2Config.from_file(relative_path("configs/qubit_cfg_1.yaml"))

    assert cfg.T1 == 1e6
    assert cfg.T2 == 3e6


def test_qubit_config():
    noise_cfg = QubitT1T2Config(T1=1e6, T2=3e6)
    cfg = QubitConfig(
        is_communication=True,
        noise_config_cls="QubitT1T2Config",
        noise_config=noise_cfg,
    )

    assert cfg.is_communication
    assert cfg.to_is_communication()
    assert cfg.to_error_model() == T1T2NoiseModel
    assert cfg.to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}


def test_qubit_config_perfect():
    for comm in [True, False]:
        cfg = QubitConfig.perfect_config(is_communication=comm)
        assert cfg.is_communication == comm
        assert cfg.to_is_communication() == comm
        assert cfg.to_error_model() == T1T2NoiseModel
        assert cfg.to_error_model_kwargs() == {"T1": 0, "T2": 0}


def test_qubit_config_file():
    cfg = QubitConfig.from_file(relative_path("configs/qubit_cfg_2.yaml"))

    assert cfg.is_communication
    assert cfg.to_is_communication()
    assert cfg.to_error_model() == T1T2NoiseModel
    assert cfg.to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}


def test_gate_depolarise_config():
    cfg = GateDepolariseConfig(duration=4e3, depolarise_prob=0.2)

    assert cfg.duration == 4e3
    assert cfg.depolarise_prob == 0.2


def test_gate_depolarise_config_file():
    cfg = GateDepolariseConfig.from_file(relative_path("configs/gate_cfg_1.yaml"))

    assert cfg.duration == 4e3
    assert cfg.depolarise_prob == 0.2
    assert cfg.to_duration() == 4e3
    assert cfg.to_error_model() == DepolarNoiseModel
    assert cfg.to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_gate_config():
    noise_cfg = GateDepolariseConfig(duration=4e3, depolarise_prob=0.2)
    cfg = GateConfig(
        name="INSTR_X", noise_config_cls="GateDepolariseConfig", noise_config=noise_cfg
    )

    assert cfg.name == "INSTR_X"
    assert cfg.to_instruction() == INSTR_X
    assert cfg.to_duration() == 4e3
    assert cfg.to_error_model() == DepolarNoiseModel
    assert cfg.to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_gate_config_perfect():
    cfg = GateConfig.perfect_config(name="INSTR_X", duration=4e3)

    assert cfg.name == "INSTR_X"
    assert cfg.to_instruction() == INSTR_X
    assert cfg.to_duration() == 4e3
    assert cfg.to_error_model() == DepolarNoiseModel
    assert cfg.to_error_model_kwargs() == {
        "depolar_rate": 0,
        "time_independent": True,
    }


def test_gate_config_file():
    cfg = GateConfig.from_file(relative_path("configs/gate_cfg_2.yaml"))

    assert cfg.name == "INSTR_X"
    assert cfg.to_instruction() == INSTR_X
    assert cfg.to_duration() == 4e3
    assert cfg.to_error_model() == DepolarNoiseModel
    assert cfg.to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_topology_config():
    qubit_noise_cfg = QubitT1T2Config(T1=1e6, T2=3e6)
    qubit_cfg = QubitConfig(
        is_communication=True,
        noise_config_cls="QubitT1T2Config",
        noise_config=qubit_noise_cfg,
    )
    gate_noise_cfg = GateDepolariseConfig(duration=4e3, depolarise_prob=0.2)
    gate_cfg = GateConfig(
        name="INSTR_X",
        noise_config_cls="GateDepolariseConfig",
        noise_config=gate_noise_cfg,
    )

    cfg = TopologyConfig(
        qubits=[QubitIdConfig(qubit_id=0, qubit_config=qubit_cfg)],
        single_gates=[SingleGateConfig(qubit_id=0, gate_configs=[gate_cfg])],
        multi_gates=[],
    )

    assert cfg.qubits[0].qubit_id == 0
    assert cfg.qubits[0].qubit_config.to_is_communication()
    assert cfg.qubits[0].qubit_config.to_error_model() == T1T2NoiseModel
    assert cfg.qubits[0].qubit_config.to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}
    assert cfg.single_gates[0].qubit_id == 0
    assert cfg.single_gates[0].gate_configs[0].to_instruction() == INSTR_X
    assert cfg.single_gates[0].gate_configs[0].to_duration() == 4e3
    assert cfg.single_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.single_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }

    # check interface
    assert cfg.get_qubit_configs()[0].to_is_communication()
    assert cfg.get_qubit_configs()[0].to_error_model() == T1T2NoiseModel
    assert cfg.get_qubit_configs()[0].to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}
    assert cfg.get_single_gate_configs()[0][0].to_instruction() == INSTR_X
    assert cfg.get_single_gate_configs()[0][0].to_duration() == 4e3
    assert cfg.get_single_gate_configs()[0][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_single_gate_configs()[0][0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_topology_config_perfect_uniform():
    cfg = TopologyConfig.perfect_config_uniform(
        num_qubits=1,
        single_instructions=["INSTR_X"],
        single_duration=4e3,
        two_instructions=[],
        two_duration=0,
    )

    assert cfg.qubits[0].qubit_id == 0
    assert cfg.qubits[0].qubit_config.to_is_communication()
    assert cfg.qubits[0].qubit_config.to_error_model() == T1T2NoiseModel
    assert cfg.qubits[0].qubit_config.to_error_model_kwargs() == {"T1": 0, "T2": 0}
    assert cfg.single_gates[0].qubit_id == 0
    assert cfg.single_gates[0].gate_configs[0].to_instruction() == INSTR_X
    assert cfg.single_gates[0].gate_configs[0].to_duration() == 4e3
    assert cfg.single_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.single_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0,
        "time_independent": True,
    }

    # check interface
    assert cfg.get_qubit_configs()[0].to_is_communication()
    assert cfg.get_qubit_configs()[0].to_error_model() == T1T2NoiseModel
    assert cfg.get_qubit_configs()[0].to_error_model_kwargs() == {"T1": 0, "T2": 0}
    assert cfg.get_single_gate_configs()[0][0].to_instruction() == INSTR_X
    assert cfg.get_single_gate_configs()[0][0].to_duration() == 4e3
    assert cfg.get_single_gate_configs()[0][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_single_gate_configs()[0][0].to_error_model_kwargs() == {
        "depolar_rate": 0,
        "time_independent": True,
    }


def test_topology_config_file():
    cfg = TopologyConfig.from_file(relative_path("configs/topology_cfg_1.yaml"))

    assert cfg.qubits[0].qubit_id == 0
    assert cfg.qubits[0].qubit_config.to_is_communication()
    assert cfg.qubits[0].qubit_config.to_error_model() == T1T2NoiseModel
    assert cfg.qubits[0].qubit_config.to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}
    assert cfg.single_gates[0].qubit_id == 0
    assert cfg.single_gates[0].gate_configs[0].to_instruction() == INSTR_X
    assert cfg.single_gates[0].gate_configs[0].to_duration() == 4e3
    assert cfg.single_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.single_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }

    # check interface
    assert cfg.get_qubit_configs()[0].to_is_communication()
    assert cfg.get_qubit_configs()[0].to_error_model() == T1T2NoiseModel
    assert cfg.get_qubit_configs()[0].to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}
    assert cfg.get_single_gate_configs()[0][0].to_instruction() == INSTR_X
    assert cfg.get_single_gate_configs()[0][0].to_duration() == 4e3
    assert cfg.get_single_gate_configs()[0][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_single_gate_configs()[0][0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_topology_config_file_2():
    cfg = TopologyConfig.from_file(relative_path("configs/topology_cfg_2.yaml"))

    assert cfg.qubits[0].qubit_id == 0
    assert cfg.qubits[0].qubit_config.to_is_communication()
    assert cfg.qubits[0].qubit_config.to_error_model() == T1T2NoiseModel
    assert cfg.qubits[0].qubit_config.to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}
    assert cfg.qubits[1].qubit_id == 1
    assert not cfg.qubits[1].qubit_config.to_is_communication()
    assert cfg.qubits[1].qubit_config.to_error_model() == T1T2NoiseModel
    assert cfg.qubits[1].qubit_config.to_error_model_kwargs() == {"T1": 2e6, "T2": 4e6}
    assert cfg.single_gates[0].qubit_id == 0
    assert cfg.single_gates[0].gate_configs[0].to_instruction() == INSTR_X
    assert cfg.single_gates[0].gate_configs[0].to_duration() == 2e3
    assert cfg.single_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.single_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }
    assert cfg.single_gates[0].gate_configs[1].to_instruction() == INSTR_Y
    assert cfg.single_gates[0].gate_configs[1].to_duration() == 4e3
    assert cfg.single_gates[0].gate_configs[1].to_error_model() == DepolarNoiseModel
    assert cfg.single_gates[0].gate_configs[1].to_error_model_kwargs() == {
        "depolar_rate": 0.4,
        "time_independent": True,
    }
    assert cfg.single_gates[1].qubit_id == 1
    assert cfg.single_gates[1].gate_configs[0].to_instruction() == INSTR_Z
    assert cfg.single_gates[1].gate_configs[0].to_duration() == 6e3
    assert cfg.single_gates[1].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.single_gates[1].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0.6,
        "time_independent": True,
    }

    # check interface
    assert cfg.get_qubit_configs()[0].to_is_communication()
    assert cfg.get_qubit_configs()[0].to_error_model() == T1T2NoiseModel
    assert cfg.get_qubit_configs()[0].to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6}
    assert not cfg.get_qubit_configs()[1].to_is_communication()
    assert cfg.get_qubit_configs()[1].to_error_model() == T1T2NoiseModel
    assert cfg.get_qubit_configs()[1].to_error_model_kwargs() == {"T1": 2e6, "T2": 4e6}
    assert cfg.get_single_gate_configs()[0][0].to_instruction() == INSTR_X
    assert cfg.get_single_gate_configs()[0][0].to_duration() == 2e3
    assert cfg.get_single_gate_configs()[0][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_single_gate_configs()[0][0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }
    assert cfg.get_single_gate_configs()[0][1].to_instruction() == INSTR_Y
    assert cfg.get_single_gate_configs()[0][1].to_duration() == 4e3
    assert cfg.get_single_gate_configs()[0][1].to_error_model() == DepolarNoiseModel
    assert cfg.get_single_gate_configs()[0][1].to_error_model_kwargs() == {
        "depolar_rate": 0.4,
        "time_independent": True,
    }
    assert cfg.get_single_gate_configs()[1][0].to_instruction() == INSTR_Z
    assert cfg.get_single_gate_configs()[1][0].to_duration() == 6e3
    assert cfg.get_single_gate_configs()[1][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_single_gate_configs()[1][0].to_error_model_kwargs() == {
        "depolar_rate": 0.6,
        "time_independent": True,
    }


def test_topology_config_multi_gate():
    qubit_noise_cfg = QubitT1T2Config(T1=1e6, T2=3e6)
    qubit_cfg = QubitConfig(
        is_communication=True,
        noise_config_cls="QubitT1T2Config",
        noise_config=qubit_noise_cfg,
    )
    gate_noise_cfg = GateDepolariseConfig(duration=4e3, depolarise_prob=0.2)
    gate_cfg = GateConfig(
        name="INSTR_CNOT",
        noise_config_cls="GateDepolariseConfig",
        noise_config=gate_noise_cfg,
    )

    cfg = TopologyConfig(
        qubits=[
            QubitIdConfig(qubit_id=0, qubit_config=qubit_cfg),
            QubitIdConfig(qubit_id=1, qubit_config=qubit_cfg),
        ],
        single_gates=[],
        multi_gates=[MultiGateConfig(qubit_ids=[0, 1], gate_configs=[gate_cfg])],
    )

    for i in [0, 1]:
        assert cfg.qubits[i].qubit_id == i
        assert cfg.qubits[i].qubit_config.to_is_communication()
        assert cfg.qubits[i].qubit_config.to_error_model() == T1T2NoiseModel
        assert cfg.qubits[i].qubit_config.to_error_model_kwargs() == {
            "T1": 1e6,
            "T2": 3e6,
        }

    assert cfg.multi_gates[0].qubit_ids == [0, 1]
    assert cfg.multi_gates[0].gate_configs[0].to_instruction() == INSTR_CNOT
    assert cfg.multi_gates[0].gate_configs[0].to_duration() == 4e3
    assert cfg.multi_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.multi_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }

    # check interface
    for i in [0, 1]:
        assert cfg.get_qubit_configs()[i].to_is_communication()
        assert cfg.get_qubit_configs()[i].to_error_model() == T1T2NoiseModel
        assert cfg.get_qubit_configs()[i].to_error_model_kwargs() == {
            "T1": 1e6,
            "T2": 3e6,
        }
    q01 = MultiQubit([0, 1])
    assert cfg.get_multi_gate_configs()[q01][0].to_instruction() == INSTR_CNOT
    assert cfg.get_multi_gate_configs()[q01][0].to_duration() == 4e3
    assert cfg.get_multi_gate_configs()[q01][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_multi_gate_configs()[q01][0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_topology_config_multi_gate_perfect_uniform():
    cfg = TopologyConfig.perfect_config_uniform(
        num_qubits=2,
        single_instructions=[],
        single_duration=0,
        two_instructions=["INSTR_CNOT"],
        two_duration=4e3,
    )

    for i in [0, 1]:
        assert cfg.qubits[i].qubit_id == i
        assert cfg.qubits[i].qubit_config.to_is_communication()
        assert cfg.qubits[i].qubit_config.to_error_model() == T1T2NoiseModel
        assert cfg.qubits[i].qubit_config.to_error_model_kwargs() == {
            "T1": 0,
            "T2": 0,
        }

    assert cfg.multi_gates[0].qubit_ids == [0, 1]
    assert cfg.multi_gates[0].gate_configs[0].to_instruction() == INSTR_CNOT
    assert cfg.multi_gates[0].gate_configs[0].to_duration() == 4e3
    assert cfg.multi_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.multi_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0,
        "time_independent": True,
    }

    # check interface
    for i in [0, 1]:
        assert cfg.get_qubit_configs()[i].to_is_communication()
        assert cfg.get_qubit_configs()[i].to_error_model() == T1T2NoiseModel
        assert cfg.get_qubit_configs()[i].to_error_model_kwargs() == {
            "T1": 0,
            "T2": 0,
        }
    q01 = MultiQubit([0, 1])
    assert cfg.get_multi_gate_configs()[q01][0].to_instruction() == INSTR_CNOT
    assert cfg.get_multi_gate_configs()[q01][0].to_duration() == 4e3
    assert cfg.get_multi_gate_configs()[q01][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_multi_gate_configs()[q01][0].to_error_model_kwargs() == {
        "depolar_rate": 0,
        "time_independent": True,
    }


def test_topology_config_multi_gate_perfect_star():
    cfg = TopologyConfig.perfect_config_star(
        num_qubits=3,
        comm_instructions=["INSTR_X"],
        comm_duration=5e3,
        mem_instructions=["INSTR_Y", "INSTR_Z"],
        mem_duration=10e3,
        two_instructions=["INSTR_CNOT"],
        two_duration=200e3,
    )

    for i, comm in zip([0, 1, 2], [True, False, False]):
        assert cfg.qubits[i].qubit_id == i
        assert cfg.qubits[i].qubit_config.to_is_communication() == comm
        assert cfg.qubits[i].qubit_config.to_error_model() == T1T2NoiseModel
        assert cfg.qubits[i].qubit_config.to_error_model_kwargs() == {
            "T1": 0,
            "T2": 0,
        }

    assert cfg.single_gates[0].qubit_id == 0
    assert len(cfg.single_gates[0].gate_configs) == 1
    assert cfg.single_gates[0].gate_configs[0].to_instruction() == INSTR_X
    assert cfg.single_gates[0].gate_configs[0].to_duration() == 5e3
    assert cfg.single_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.single_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0,
        "time_independent": True,
    }

    for i in [1, 2]:
        assert cfg.single_gates[i].qubit_id == i
        assert len(cfg.single_gates[i].gate_configs) == 2
        assert cfg.single_gates[i].gate_configs[0].to_instruction() == INSTR_Y
        assert cfg.single_gates[i].gate_configs[1].to_instruction() == INSTR_Z
        for j in [0, 1]:
            assert cfg.single_gates[i].gate_configs[j].to_duration() == 10e3
            assert (
                cfg.single_gates[i].gate_configs[j].to_error_model()
                == DepolarNoiseModel
            )
            assert cfg.single_gates[i].gate_configs[j].to_error_model_kwargs() == {
                "depolar_rate": 0,
                "time_independent": True,
            }

    assert len(cfg.multi_gates) == 2
    for i in [1, 2]:
        assert cfg.multi_gates[i - 1].qubit_ids == [0, i]
        assert len(cfg.multi_gates[i - 1].gate_configs) == 1
        assert len(cfg.multi_gates[i - 1].gate_configs) == 1
        assert cfg.multi_gates[i - 1].gate_configs[0].to_instruction() == INSTR_CNOT
        assert cfg.multi_gates[i - 1].gate_configs[0].to_duration() == 200e3
        assert (
            cfg.multi_gates[i - 1].gate_configs[0].to_error_model() == DepolarNoiseModel
        )
        assert cfg.multi_gates[i - 1].gate_configs[0].to_error_model_kwargs() == {
            "depolar_rate": 0,
            "time_independent": True,
        }


def test_topology_config_file_multi_gate():
    cfg = TopologyConfig.from_file(relative_path("configs/topology_cfg_3.yaml"))

    for i in [0, 1]:
        assert cfg.qubits[i].qubit_id == i
        assert cfg.qubits[i].qubit_config.to_is_communication()
        assert cfg.qubits[i].qubit_config.to_error_model() == T1T2NoiseModel
        assert cfg.qubits[i].qubit_config.to_error_model_kwargs() == {
            "T1": 1e6,
            "T2": 3e6,
        }

    assert cfg.multi_gates[0].qubit_ids == [0, 1]
    assert cfg.multi_gates[0].gate_configs[0].to_instruction() == INSTR_CNOT
    assert cfg.multi_gates[0].gate_configs[0].to_duration() == 4e3
    assert cfg.multi_gates[0].gate_configs[0].to_error_model() == DepolarNoiseModel
    assert cfg.multi_gates[0].gate_configs[0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }

    # check interface
    for i in [0, 1]:
        assert cfg.get_qubit_configs()[i].to_is_communication()
        assert cfg.get_qubit_configs()[i].to_error_model() == T1T2NoiseModel
        assert cfg.get_qubit_configs()[i].to_error_model_kwargs() == {
            "T1": 1e6,
            "T2": 3e6,
        }
    q01 = MultiQubit([0, 1])
    assert cfg.get_multi_gate_configs()[q01][0].to_instruction() == INSTR_CNOT
    assert cfg.get_multi_gate_configs()[q01][0].to_duration() == 4e3
    assert cfg.get_multi_gate_configs()[q01][0].to_error_model() == DepolarNoiseModel
    assert cfg.get_multi_gate_configs()[q01][0].to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_topology_config_file_reuse_gate_def():
    cfg = TopologyConfig.from_file(relative_path("configs/topology_cfg_5.yaml"))

    assert cfg.get_single_gate_configs()[0][0].to_instruction() == INSTR_X
    assert cfg.get_single_gate_configs()[0][1].to_instruction() == INSTR_Y
    assert cfg.get_single_gate_configs()[1][0].to_instruction() == INSTR_X


def test_qubit_config_file_registry():
    class QubitT1T2T3Config(QubitT1T2Config):
        T3: int

        @classmethod
        def from_dict(cls, dict: Any) -> QubitT1T2T3Config:
            return QubitT1T2T3Config(**dict)

        def to_error_model_kwargs(self) -> Dict[str, Any]:
            return {"T1": self.T1, "T2": self.T2, "T3": self.T3}

    class CustomQubitConfigRegistry(QubitConfigRegistry):
        @classmethod
        def map(cls) -> Dict[str, QubitNoiseConfigInterface]:
            return {"QubitT1T2T3Config": QubitT1T2T3Config}

    cfg = QubitConfig.from_file(
        relative_path("configs/qubit_cfg_2_custom_cls.yaml"),
        registry=[CustomQubitConfigRegistry],
    )

    assert cfg.is_communication
    assert cfg.to_is_communication()
    assert cfg.to_error_model() == T1T2NoiseModel
    assert cfg.to_error_model_kwargs() == {"T1": 1e6, "T2": 3e6, "T3": 5e6}


def test_gate_config_file_registry():
    class MyLeetNoise(GateNoiseConfigInterface, BaseModel):
        my_noise_param: int

        @classmethod
        def from_dict(cls, dict: Any) -> MyLeetNoise:
            return MyLeetNoise(**dict)

        def to_duration(self) -> int:
            return 42

        def to_error_model(self) -> Type[QuantumErrorModel]:
            return QuantumErrorModel

        def to_error_model_kwargs(self) -> Dict[str, Any]:
            return {"my_noise_param": self.my_noise_param}

    class CustomGateConfigRegistry(GateConfigRegistry):
        @classmethod
        def map(cls) -> Dict[str, GateNoiseConfigInterface]:
            return {"MyLeetNoise": MyLeetNoise}

    cfg = GateConfig.from_file(
        relative_path("configs/gate_cfg_2_custom_cls.yaml"),
        registry=[CustomGateConfigRegistry],
    )

    assert cfg.name == "INSTR_X"
    assert cfg.to_instruction() == INSTR_X
    assert cfg.to_duration() == 42
    assert cfg.to_error_model() == QuantumErrorModel
    assert cfg.to_error_model_kwargs() == {
        "my_noise_param": 1337,
    }


def test_custom_instruction():
    class CustomInstrRegistry(InstrConfigRegistry):
        @classmethod
        def map(cls) -> Dict[str, NetSquidInstruction]:
            return {"MY_CUSTOM_INSTR": INSTR_H}

    cfg = GateConfig.from_file(relative_path("configs/gate_cfg_2_custom_instr.yaml"))

    assert cfg.name == "MY_CUSTOM_INSTR"
    assert cfg.to_instruction(registry=[CustomInstrRegistry]) == INSTR_H
    assert cfg.to_duration() == 4e3
    assert cfg.to_error_model() == DepolarNoiseModel
    assert cfg.to_error_model_kwargs() == {
        "depolar_rate": 0.2,
        "time_independent": True,
    }


def test_latencies_config_file():
    cfg = LatenciesConfig.from_file(relative_path("configs/latencies_cfg_1.yaml"))

    assert cfg.host_instr_time == 500
    assert cfg.qnos_instr_time == 2000
    assert cfg.host_peer_latency == 2e6

    # check interface
    assert cfg.get_host_instr_time() == 500
    assert cfg.get_qnos_instr_time() == 2000
    assert cfg.get_host_peer_latency() == 2e6


def test_latencies_config_file_default_values():
    cfg = LatenciesConfig.from_file(relative_path("configs/latencies_cfg_2.yaml"))

    # explicitly given by cfg file
    assert cfg.host_instr_time == 200

    # not given in the cfg file, so they should default to 0
    assert cfg.qnos_instr_time == 0
    assert cfg.host_peer_latency == 0

    # check interface
    assert cfg.get_host_instr_time() == 200
    assert cfg.get_qnos_instr_time() == 0
    assert cfg.get_host_peer_latency() == 0


def test_procnode_config_file():
    cfg = ProcNodeConfig.from_file(relative_path("configs/procnode_cfg_1.yaml"))

    # the topology used in this file is the same as in configs/topology_cfg_1.yaml
    expected_topology = TopologyConfig.from_file(
        relative_path("configs/topology_cfg_1.yaml")
    )

    assert cfg.node_name == "client_node"
    assert cfg.node_id == 2
    assert cfg.topology == expected_topology
    assert cfg.latencies.host_instr_time == 500
    assert cfg.latencies.qnos_instr_time == 2000
    assert cfg.latencies.host_peer_latency == 2e6


def test_procnode_config_file_default_values():
    cfg = ProcNodeConfig.from_file(relative_path("configs/procnode_cfg_2.yaml"))

    # following 2 items are not given in the cfg file, so they should default to 0
    assert cfg.latencies.host_instr_time == 0
    assert cfg.latencies.qnos_instr_time == 0

    # explicitly given by cfg file
    assert cfg.latencies.host_peer_latency == 2e6


def test_perfect_sampler_config():
    cfg = PerfectSamplerConfig(cycle_time=10)

    assert cfg.to_sampler_factory() == PerfectStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {"cycle_time": 10}


def test_depolarise_sampler_config():
    cfg = DepolariseSamplerConfig(cycle_time=10, prob_max_mixed=0.3, prob_success=0.1)

    assert cfg.to_sampler_factory() == DepolariseWithFailureStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {
        "cycle_time": 10,
        "prob_max_mixed": 0.3,
        "prob_success": 0.1,
    }


def test_link_config():
    sampler_cfg = PerfectSamplerConfig(cycle_time=10)
    cfg = LinkConfig(
        state_delay=500,
        sampler_config_cls="PerfectSamplerConfig",
        sampler_config=sampler_cfg,
    )

    assert cfg.state_delay == 500
    assert cfg.to_state_delay() == 500
    assert cfg.to_sampler_factory() == PerfectStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {"cycle_time": 10}


def test_link_config_2():
    sampler_cfg = DepolariseSamplerConfig(
        cycle_time=10, prob_max_mixed=0.3, prob_success=0.1
    )
    cfg = LinkConfig(
        state_delay=500,
        sampler_config_cls="DepolariseSamplerConfig",
        sampler_config=sampler_cfg,
    )

    assert cfg.state_delay == 500
    assert cfg.to_state_delay() == 500
    assert cfg.to_sampler_factory() == DepolariseWithFailureStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {
        "cycle_time": 10,
        "prob_max_mixed": 0.3,
        "prob_success": 0.1,
    }


def test_link_config_perfect():
    cfg = LinkConfig.perfect_config(200)

    assert cfg.state_delay == 200
    assert cfg.to_state_delay() == 200
    assert cfg.to_sampler_factory() == PerfectStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {"cycle_time": 0}


def test_link_config_depolarise():
    cfg = LinkConfig.depolarise_config(fidelity=0.8, state_delay=200)

    assert cfg.to_state_delay() == 200
    assert cfg.to_sampler_factory() == DepolariseWithFailureStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {
        "cycle_time": 0,
        "prob_max_mixed": fidelity_to_prob_max_mixed(0.8),
        "prob_success": 1,
    }


def test_link_config_from_file():
    cfg = LinkConfig.from_file(relative_path("configs/link_cfg_1.yaml"))

    assert cfg.state_delay == 750
    assert cfg.to_state_delay() == 750
    assert cfg.to_sampler_factory() == PerfectStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {"cycle_time": 25}


def test_link_config_from_file_2():
    cfg = LinkConfig.from_file(relative_path("configs/link_cfg_2.yaml"))

    assert cfg.state_delay == 750
    assert cfg.to_state_delay() == 750
    assert cfg.to_sampler_factory() == DepolariseWithFailureStateSamplerFactory
    assert cfg.to_sampler_kwargs() == {
        "cycle_time": 10,
        "prob_max_mixed": 0.3,
        "prob_success": 0.1,
    }


def test_link_between_nodes_from_file():
    cfg = LinkBetweenNodesConfig.from_file(relative_path("configs/link_cfg_3.yaml"))

    assert cfg.node_id1 == 2
    assert cfg.node_id2 == 5
    assert cfg.link_config.state_delay == 750
    assert cfg.link_config.to_state_delay() == 750
    assert (
        cfg.link_config.to_sampler_factory() == DepolariseWithFailureStateSamplerFactory
    )
    assert cfg.link_config.to_sampler_kwargs() == {
        "cycle_time": 10,
        "prob_max_mixed": 0.3,
        "prob_success": 0.1,
    }


if __name__ == "__main__":
    test_qubit_t1t2_config()
    test_qubit_t1t2_config_file()
    test_qubit_config()
    test_qubit_config_perfect()
    test_qubit_config_file()
    test_gate_depolarise_config()
    test_gate_depolarise_config_file()
    test_gate_config()
    test_gate_config_perfect()
    test_gate_config_file()
    test_topology_config()
    test_topology_config_perfect_uniform()
    test_topology_config_file()
    test_topology_config_file_2()
    test_topology_config_multi_gate()
    test_topology_config_file_multi_gate()
    test_topology_config_multi_gate_perfect_uniform()
    test_topology_config_multi_gate_perfect_star()
    test_topology_config_file_reuse_gate_def()
    test_qubit_config_file_registry()
    test_gate_config_file_registry()
    test_custom_instruction()
    test_latencies_config_file()
    test_latencies_config_file_default_values()
    test_procnode_config_file()
    test_procnode_config_file_default_values()
    test_perfect_sampler_config()
    test_depolarise_sampler_config()
    test_link_config()
    test_link_config_2()
    test_link_config_perfect()
    test_link_config_depolarise()
    test_link_config_from_file()
    test_link_config_from_file_2()
    test_link_between_nodes_from_file()



from __future__ import annotations

import logging
from typing import Generator, List

from netqasm.lang.operand import Template

from pydynaa import EventExpression
from qoala.lang import hostlang
from qoala.lang.request import CallbackType
from qoala.runtime.message import LrCallTuple, Message, RrCallTuple
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.host.hostinterface import HostInterface, HostLatencies
from qoala.sim.process import QoalaProcess
from qoala.util.logging import LogManager


class HostProcessor:
    """Does not have state itself. Acts on and changes process objects."""

    def __init__(
        self,
        interface: HostInterface,
        latencies: HostLatencies,
        asynchronous: bool = False,
    ) -> None:
        self._interface = interface
        self._latencies = latencies
        self._asynchronous = asynchronous

        # TODO: name
        self._name = f"{interface.name}_HostProcessor"
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

    def initialize(self, process: QoalaProcess) -> None:
        host_mem = process.prog_memory.host_mem
        inputs = process.prog_instance.inputs
        for name, value in inputs.values.items():
            host_mem.write(name, value)

    def assign_instr_index(
        self, process: QoalaProcess, instr_idx: int
    ) -> Generator[EventExpression, None, None]:
        program = process.prog_instance.program
        instr = program.instructions[instr_idx]
        yield from self.assign_instr(process, instr)

    def assign_block(
        self, process: QoalaProcess, block_name: str
    ) -> Generator[EventExpression, None, None]:
        block = process.program.get_block(block_name)
        for instr in block.instructions:
            yield from self.assign_instr(process, instr)

    def assign_instr(
        self, process: QoalaProcess, instr: hostlang.ClassicalIqoalaOp
    ) -> Generator[EventExpression, None, None]:
        csockets = process.csockets
        host_mem = process.prog_memory.host_mem

        # Instruction duration is simulated for each instruction by adding a "wait".
        # Duration of wait is "host_instr_time".
        # Half of it is applied *before* any operations.
        # The other half is applied *after* all 'reads' from shared memory,
        # and *before* any 'writes' to shared memory.
        # See #29 for rationale.

        # Apply half of the instruction duration.
        instr_time = self._latencies.host_instr_time
        first_half = instr_time / 2
        second_half = instr_time - first_half  # just to make it adds up

        self._logger.info(f"Interpreting LHR instruction {instr}")
        if isinstance(instr, hostlang.AssignCValueOp):
            yield from self._interface.wait(first_half)
            value = instr.attributes[0]
            assert isinstance(value, int)
            loc = instr.results[0]  # type: ignore
            self._logger.info(f"writing {value} to {loc}")
            yield from self._interface.wait(second_half)
            host_mem.write(loc, value)
        elif isinstance(instr, hostlang.SendCMsgOp):
            assert isinstance(instr.arguments[0], str)
            assert isinstance(instr.arguments[1], str)

            csck_id = host_mem.read(instr.arguments[0])
            csck = csockets[csck_id]
            value = host_mem.read(instr.arguments[1])
            self._logger.info(f"sending msg {value}")
            csck.send_int(value)
            # Simulate instruction duration.
            yield from self._interface.wait(self._latencies.host_instr_time)
        elif isinstance(instr, hostlang.ReceiveCMsgOp):
            assert isinstance(instr.arguments[0], str)
            assert isinstance(instr.results, list)
            csck_id = host_mem.read(instr.arguments[0])
            csck = csockets[csck_id]
            msg = yield from csck.recv_int()

            yield from self._interface.wait(self._latencies.host_peer_latency)
            host_mem.write(instr.results[0], msg)
            self._logger.info(f"received msg {msg}")
        elif isinstance(instr, hostlang.AddCValueOp):
            yield from self._interface.wait(first_half)
            assert isinstance(instr.arguments[0], str)
            assert isinstance(instr.arguments[1], str)
            arg0 = host_mem.read(instr.arguments[0])
            arg1 = host_mem.read(instr.arguments[1])
            loc = instr.results[0]  # type: ignore
            result = arg0 + arg1
            self._logger.info(f"computing {loc} = {arg0} + {arg1} = {result}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            host_mem.write(loc, result)
        elif isinstance(instr, hostlang.MultiplyConstantCValueOp):
            yield from self._interface.wait(first_half)
            assert isinstance(instr.arguments[0], str)
            arg0 = host_mem.read(instr.arguments[0])
            const = instr.attributes[0]
            assert isinstance(const, int)
            loc = instr.results[0]  # type: ignore
            result = arg0 * const
            self._logger.info(f"computing {loc} = {arg0} * {const} = {result}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            host_mem.write(loc, result)
        elif isinstance(instr, hostlang.BitConditionalMultiplyConstantCValueOp):
            yield from self._interface.wait(first_half)
            assert isinstance(instr.arguments[0], str)
            assert isinstance(instr.arguments[1], str)
            arg0 = host_mem.read(instr.arguments[0])
            cond = host_mem.read(instr.arguments[1])
            const = instr.attributes[0]
            assert isinstance(const, int)
            loc = instr.results[0]  # type: ignore
            if cond == 1:
                result = arg0 * const
            else:
                result = arg0
            self._logger.info(f"computing {loc} = {arg0} * {const}^{cond} = {result}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            host_mem.write(loc, result)
        elif isinstance(instr, hostlang.RunSubroutineOp):
            lrcall = self.prepare_lr_call(process, instr)

            # Send a message to Qnos asking to run the routine.
            self._interface.send_qnos_msg(Message(lrcall))

            # Wait until Qnos says it has finished.
            yield from self._interface.receive_qnos_msg()

            self.post_lr_call(process, instr, lrcall)
        elif isinstance(instr, hostlang.RunRequestOp):
            yield from self._interface.wait(first_half)
            rrcall = self.prepare_rr_call(process, instr)

            # Send a message to the Netstack asking to run the routine.
            self._interface.send_netstack_msg(Message(rrcall))

            # Wait until the Netstack says it has finished.
            yield from self._interface.receive_netstack_msg()

            # TODO: read results

        elif isinstance(instr, hostlang.ReturnResultOp):
            assert isinstance(instr.arguments[0], str)
            loc = instr.arguments[0]
            value = host_mem.read(loc)
            self._logger.info(f"returning {loc} = {value}")
            # Simulate instruction duration.
            yield from self._interface.wait(second_half)
            process.result.values[loc] = value

    def prepare_lr_call(
        self, process: QoalaProcess, instr: hostlang.RunSubroutineOp
    ) -> LrCallTuple:
        host_mem = process.prog_memory.host_mem

        assert isinstance(instr.arguments[0], hostlang.IqoalaVector)
        arg_vec: hostlang.IqoalaVector = instr.arguments[0]
        args = arg_vec.values
        subrt_name = instr.attributes[0]
        assert isinstance(subrt_name, str)

        routine = process.get_local_routine(subrt_name)
        self._logger.info(f"executing subroutine {routine}")

        arg_values = {arg: host_mem.read(arg) for arg in args}

        # self._logger.info(f"instantiating subroutine with values {arg_values}")
        # process.instantiate_routine(subrt_name, arg_values)

        shared_mem = process.prog_memory.shared_mem

        # Allocate input memory and write args to it.
        input_addr = shared_mem.allocate_lr_in(len(arg_values))
        shared_mem.write_lr_in(input_addr, list(arg_values.values()))

        # Allocate result memory.
        result_addr = shared_mem.allocate_lr_out(len(routine.return_vars))

        return LrCallTuple(subrt_name, input_addr, result_addr)

    def post_lr_call(
        self,
        process: QoalaProcess,
        instr: hostlang.RunSubroutineOp,
        lrcall: LrCallTuple,
    ) -> None:
        shared_mem = process.prog_memory.shared_mem

        # Collect the host variables that should obtain the LR results.
        result_vars: List[str]
        if isinstance(instr.results, list):
            result_vars = instr.results
        elif isinstance(instr.results, hostlang.IqoalaVector):
            result_vec: hostlang.IqoalaVector = instr.results
            result_vars = result_vec.values
        else:
            raise RuntimeError

        # Read the results from shared memory.
        result = shared_mem.read_lr_out(lrcall.result_addr, len(result_vars))

        # Copy results to local host variables.
        assert len(result) == len(result_vars)
        for value, var in zip(result, result_vars):
            process.host_mem.write(var, value)

    def prepare_rr_call(
        self, process: QoalaProcess, instr: hostlang.RunRequestOp
    ) -> RrCallTuple:
        host_mem = process.prog_memory.host_mem

        assert isinstance(instr.arguments[0], hostlang.IqoalaVector)
        arg_vec: hostlang.IqoalaVector = instr.arguments[0]
        args = arg_vec.values
        routine_name = instr.attributes[0]
        assert isinstance(routine_name, str)

        routine = process.get_request_routine(routine_name)
        self._logger.info(f"executing request routine {routine}")

        arg_values = {arg: host_mem.read(arg) for arg in args}

        shared_mem = process.prog_memory.shared_mem

        # Allocate input memory for RR itself and write args to it.
        input_addr = shared_mem.allocate_rr_in(len(arg_values))
        shared_mem.write_rr_in(input_addr, list(arg_values.values()))

        cb_input_addrs: List[MemAddr] = []
        cb_output_addrs: List[MemAddr] = []

        # TODO: refactor!!
        # the `num_pairs` entry of an RR may be a template.
        # Its value should be provided in the ProgramInput of this ProgamInstance.
        # This value is filled in when `instantiating` the RR, but currently this only
        # happens by the NetstackProcessor when it's assigned to execute the RR.
        # The filled-in value is then part of a `RunningRequestRoutine`. However, it is
        # not accessible by this code here.
        # For now we use the following 'hack' where we peek in the ProgramInputs:
        if isinstance(routine.request.num_pairs, Template):
            template_name = routine.request.num_pairs.name
            num_pairs = process.prog_instance.inputs.values[template_name]
        else:
            num_pairs = routine.request.num_pairs

        # Allocate memory for callbacks.
        if routine.callback is not None:
            if routine.callback_type == CallbackType.SEQUENTIAL:
                cb_routine = process.get_local_routine(routine.callback)
                for _ in range(num_pairs):
                    # Allocate input memory.
                    cb_args = cb_routine.subroutine.arguments
                    # TODO: can it just be LR_in instead of CR_in?
                    cb_input_addrs.append(shared_mem.allocate_cr_in(len(cb_args)))

                    # Allocate result memory.
                    cb_results = cb_routine.return_vars
                    cb_output_addrs.append(shared_mem.allocate_lr_out(len(cb_results)))
            elif routine.callback_type == CallbackType.WAIT_ALL:
                cb_routine = process.get_local_routine(routine.callback)
                # Allocate input memory.
                cb_args = cb_routine.subroutine.arguments
                # TODO: can it just be LR_in instead of CR_in?
                cb_input_addrs.append(shared_mem.allocate_cr_in(len(cb_args)))

                # Allocate result memory.
                cb_results = cb_routine.return_vars
                cb_output_addrs.append(shared_mem.allocate_lr_out(len(cb_results)))

        # Allocate result memory for RR itself.
        result_addr = shared_mem.allocate_rr_out(len(routine.return_vars))

        return RrCallTuple(
            routine_name, input_addr, result_addr, cb_input_addrs, cb_output_addrs
        )

    def post_rr_call(
        self, process: QoalaProcess, instr: hostlang.RunRequestOp, rrcall: RrCallTuple
    ) -> None:
        shared_mem = process.prog_memory.shared_mem
        routine = process.get_request_routine(rrcall.routine_name)

        # Read the RR results from shared memory.
        rr_result = shared_mem.read_rr_out(rrcall.result_addr, len(routine.return_vars))

        # Bit of a hack; see prepare_rr_call comments.
        if isinstance(routine.request.num_pairs, Template):
            template_name = routine.request.num_pairs.name
            num_pairs = process.prog_instance.inputs.values[template_name]
        else:
            num_pairs = routine.request.num_pairs

        # Read the callback results from shared memory.
        cb_results: List[int] = []
        if routine.callback is not None:
            if routine.callback_type == CallbackType.SEQUENTIAL:
                cb_routine = process.get_local_routine(routine.callback)
                for i in range(num_pairs):
                    # Read result memory.
                    cb_ret_vars = cb_routine.return_vars
                    cb_output_addr = rrcall.cb_output_addrs[i]
                    result = shared_mem.read_lr_out(cb_output_addr, len(cb_ret_vars))
                    cb_results.extend(result)
            elif routine.callback_type == CallbackType.WAIT_ALL:
                cb_routine = process.get_local_routine(routine.callback)
                # Read result memory.
                cb_ret_vars = cb_routine.return_vars
                cb_output_addr = rrcall.cb_output_addrs[0]
                result = shared_mem.read_lr_out(cb_output_addr, len(cb_ret_vars))
                cb_results.extend(result)

        all_results = rr_result + cb_results
        # At this point, `all_results` contains all the results of both the RR itself
        # as well as from all the callbacks, in order.

        # Collect the host variables to which to copy these results.
        result_vars: List[str]
        if isinstance(instr.results, list):
            result_vars = instr.results
        elif isinstance(instr.results, hostlang.IqoalaVector):
            result_vec: hostlang.IqoalaVector = instr.results
            result_vars = result_vec.values
        else:
            raise RuntimeError

        # Copy all results to local host variables.
        assert len(all_results) == len(result_vars)
        for value, var in zip(all_results, result_vars):
            process.host_mem.write(var, value)



import pytest
from netqasm.lang.instr import core, nv
from netqasm.lang.instr.flavour import NVFlavour
from netsquid.components.instructions import (
    INSTR_CXDIR,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
)

from qoala.lang.common import MultiQubit
from qoala.lang.ehi import EhiGateInfo, EhiQubitInfo
from qoala.runtime.config import DepolariseSamplerConfig, LinkConfig
from qoala.runtime.lhi import (
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiTopologyBuilder,
)
from qoala.runtime.lhi_to_ehi import LhiConverter, NvToNvInterface
from qoala.util.math import prob_max_mixed_to_fidelity


def test_topology_to_ehi():
    topology = LhiTopologyBuilder.perfect_uniform(
        num_qubits=2,
        single_instructions=[
            INSTR_INIT,
            INSTR_ROT_X,
            INSTR_ROT_Y,
            INSTR_ROT_Z,
            INSTR_MEASURE,
        ],
        single_duration=5e3,
        two_instructions=[INSTR_CXDIR],
        two_duration=100e3,
    )

    latencies = LhiLatencies(
        host_instr_time=2,
        qnos_instr_time=3,
        host_peer_latency=4,
    )

    interface = NvToNvInterface()
    ehi = LhiConverter.to_ehi(topology, interface, latencies)

    assert ehi.qubit_infos == {
        0: EhiQubitInfo(is_communication=True, decoherence_rate=0),
        1: EhiQubitInfo(is_communication=True, decoherence_rate=0),
    }

    assert ehi.flavour == NVFlavour

    single_gates = [
        EhiGateInfo(instr, 5e3, 0)
        for instr in [
            core.InitInstruction,
            nv.RotXInstruction,
            nv.RotYInstruction,
            nv.RotZInstruction,
            core.MeasInstruction,
        ]
    ]
    assert ehi.single_gate_infos == {0: single_gates, 1: single_gates}

    multi_gates = [EhiGateInfo(nv.ControlledRotXInstruction, 100e3, 0)]

    assert ehi.multi_gate_infos == {
        MultiQubit([0, 1]): multi_gates,
        MultiQubit([1, 0]): multi_gates,
    }

    assert ehi.latencies.host_instr_time == 2
    assert ehi.latencies.qnos_instr_time == 3
    assert ehi.latencies.host_peer_latency == 4


def test_link_info_to_ehi_perfect():
    cfg = LinkConfig.perfect_config(state_delay=1200)
    lhi_info = LhiLinkInfo.from_config(cfg)
    ehi_info = LhiConverter.link_info_to_ehi(lhi_info)

    assert ehi_info.duration == 1200
    assert ehi_info.fidelity == 1.0


def test_link_info_to_ehi_depolarise():
    state_delay = 500
    cycle_time = 10
    prob_max_mixed = 0.3
    prob_success = 0.1

    cfg = LinkConfig(
        state_delay=state_delay,
        sampler_config_cls="DepolariseSamplerConfig",
        sampler_config=DepolariseSamplerConfig(
            cycle_time=cycle_time,
            prob_max_mixed=prob_max_mixed,
            prob_success=prob_success,
        ),
    )
    lhi_info = LhiLinkInfo.from_config(cfg)

    ehi_info = LhiConverter.link_info_to_ehi(lhi_info)

    expected_duration = (cycle_time / prob_success) + state_delay
    expected_fidelity = prob_max_mixed_to_fidelity(prob_max_mixed)
    assert ehi_info.duration == pytest.approx(expected_duration)
    assert ehi_info.fidelity == pytest.approx(expected_fidelity)


def test_network_to_ehi():
    depolar_link = LhiLinkInfo.depolarise(
        cycle_time=10, prob_max_mixed=0.2, prob_success=0.5, state_delay=2000
    )
    perfect_link = LhiLinkInfo.perfect(1000)
    lhi_network = LhiNetworkInfo(links={(0, 1): depolar_link, (1, 3): perfect_link})

    ehi_network = LhiConverter.network_to_ehi(lhi_network)
    expected_duration_0_1 = (10 / 0.5) + 2000
    expected_fidelty_0_1 = prob_max_mixed_to_fidelity(0.2)

    ehi_link_0_1 = ehi_network.get_link(0, 1)
    assert ehi_link_0_1.duration == pytest.approx(expected_duration_0_1)
    assert ehi_link_0_1.fidelity == pytest.approx(expected_fidelty_0_1)

    ehi_link_1_3 = ehi_network.get_link(1, 3)
    assert ehi_link_1_3.duration == 1000
    assert ehi_link_1_3.fidelity == 1.0


if __name__ == "__main__":
    test_topology_to_ehi()
    test_link_info_to_ehi_perfect()
    test_link_info_to_ehi_depolarise()
    test_network_to_ehi()



import abc
from typing import Any, Dict, Optional, Tuple, Type

from netqasm.lang.instr import core, nv, vanilla
from netqasm.lang.instr.base import NetQASMInstruction
from netqasm.lang.instr.flavour import Flavour, NVFlavour, VanillaFlavour
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.instructions import Instruction as NetSquidInstruction
from netsquid.components.models.qerrormodels import (
    DepolarNoiseModel,
    QuantumErrorModel,
    T1T2NoiseModel,
)
from netsquid_magic.state_delivery_sampler import (
    DepolariseWithFailureStateSamplerFactory,
    PerfectStateSamplerFactory,
)

from qoala.lang.ehi import (
    EhiGateInfo,
    EhiLatencies,
    EhiLinkInfo,
    EhiNetworkInfo,
    EhiNodeInfo,
    EhiQubitInfo,
)
from qoala.runtime.lhi import (
    INSTR_MEASURE_INSTANT,
    LhiGateInfo,
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiQubitInfo,
    LhiTopology,
)
from qoala.util.math import prob_max_mixed_to_fidelity


class NativeToFlavourInterface(abc.ABC):
    @abc.abstractmethod
    def flavour(self) -> Type[Flavour]:
        raise NotImplementedError

    @abc.abstractmethod
    def map(self, ns_instr: Type[NetSquidInstruction]) -> Type[NetQASMInstruction]:
        """Responsiblity of implementor that return instructions are of the
        flavour returned by flavour()."""
        raise NotImplementedError


class GenericToVanillaInterface(NativeToFlavourInterface):
    _MAP: Dict[Type[NetSquidInstruction], Type[NetQASMInstruction]] = {
        INSTR_INIT: core.InitInstruction,
        INSTR_X: vanilla.GateXInstruction,
        INSTR_Y: vanilla.GateYInstruction,
        INSTR_Z: vanilla.GateZInstruction,
        INSTR_H: vanilla.GateHInstruction,
        INSTR_ROT_X: vanilla.RotXInstruction,
        INSTR_ROT_Y: vanilla.RotYInstruction,
        INSTR_ROT_Z: vanilla.RotZInstruction,
        INSTR_CNOT: vanilla.CnotInstruction,
        INSTR_CZ: vanilla.CphaseInstruction,
        INSTR_MEASURE: core.MeasInstruction,
        INSTR_MEASURE_INSTANT: core.MeasInstruction,
    }

    def flavour(self) -> Type[Flavour]:
        return VanillaFlavour  # type: ignore

    def map(self, ns_instr: Type[NetSquidInstruction]) -> Type[NetQASMInstruction]:
        """Responsiblity of implementor that return instructions are of the
        flavour returned by flavour()."""
        return self._MAP[ns_instr]


class NvToNvInterface(NativeToFlavourInterface):
    _MAP: Dict[Type[NetSquidInstruction], Type[NetQASMInstruction]] = {
        INSTR_INIT: core.InitInstruction,
        INSTR_ROT_X: nv.RotXInstruction,
        INSTR_ROT_Y: nv.RotYInstruction,
        INSTR_ROT_Z: nv.RotZInstruction,
        INSTR_CXDIR: nv.ControlledRotXInstruction,
        INSTR_CYDIR: nv.ControlledRotYInstruction,
        INSTR_MEASURE: core.MeasInstruction,
    }

    def flavour(self) -> Type[Flavour]:
        return NVFlavour  # type: ignore

    def map(self, ns_instr: Type[NetSquidInstruction]) -> Type[NetQASMInstruction]:
        """Responsiblity of implementor that return instructions are of the
        flavour returned by flavour()."""
        return self._MAP[ns_instr]


class LhiConverter:
    @classmethod
    def error_model_to_rate(
        cls, model: Type[QuantumErrorModel], model_kwargs: Dict[str, Any]
    ) -> float:
        if model == DepolarNoiseModel:
            return model_kwargs["depolar_rate"]  # type: ignore
        elif model == T1T2NoiseModel:
            # TODO use T2 somehow
            return model_kwargs["T1"]  # type: ignore
        else:
            raise RuntimeError("Unsupported LHI Error model")

    @classmethod
    def qubit_info_to_ehi(cls, info: LhiQubitInfo) -> EhiQubitInfo:
        return EhiQubitInfo(
            is_communication=info.is_communication,
            decoherence_rate=cls.error_model_to_rate(
                info.error_model, info.error_model_kwargs
            ),
        )

    @classmethod
    def gate_info_to_ehi(
        cls, info: LhiGateInfo, ntf: NativeToFlavourInterface
    ) -> EhiGateInfo:
        instr = ntf.map(info.instruction)
        duration = info.duration
        decoherence = cls.error_model_to_rate(info.error_model, info.error_model_kwargs)
        return EhiGateInfo(
            instruction=instr, duration=duration, decoherence=decoherence
        )

    @classmethod
    def to_ehi(
        cls,
        topology: LhiTopology,
        ntf: NativeToFlavourInterface,
        latencies: Optional[LhiLatencies] = None,
    ) -> EhiNodeInfo:
        if latencies is None:
            latencies = LhiLatencies.all_zero()

        qubit_infos = {
            id: cls.qubit_info_to_ehi(qi) for (id, qi) in topology.qubit_infos.items()
        }
        single_gate_infos = {
            id: [cls.gate_info_to_ehi(gi, ntf) for gi in gis]
            for (id, gis) in topology.single_gate_infos.items()
        }
        multi_gate_infos = {
            ids: [cls.gate_info_to_ehi(gi, ntf) for gi in gis]
            for (ids, gis) in topology.multi_gate_infos.items()
        }
        flavour = ntf.flavour()

        ehi_latencies = EhiLatencies(
            latencies.host_instr_time,
            latencies.qnos_instr_time,
            latencies.host_peer_latency,
        )

        return EhiNodeInfo(
            qubit_infos=qubit_infos,
            flavour=flavour,
            single_gate_infos=single_gate_infos,
            multi_gate_infos=multi_gate_infos,
            latencies=ehi_latencies,
        )

    @classmethod
    def link_info_to_ehi(cls, info: LhiLinkInfo) -> EhiLinkInfo:
        if info.sampler_factory == PerfectStateSamplerFactory:
            return EhiLinkInfo(duration=info.state_delay, fidelity=1.0)
        elif info.sampler_factory == DepolariseWithFailureStateSamplerFactory:
            expected_gen_duration = (
                info.sampler_kwargs["cycle_time"] / info.sampler_kwargs["prob_success"]
            )
            duration = expected_gen_duration + info.state_delay
            fidelity = prob_max_mixed_to_fidelity(info.sampler_kwargs["prob_max_mixed"])
            return EhiLinkInfo(duration=duration, fidelity=fidelity)
        else:
            raise NotImplementedError

    @classmethod
    def network_to_ehi(cls, info: LhiNetworkInfo) -> EhiNetworkInfo:
        links: Dict[Tuple[int, int], EhiLinkInfo] = {}
        for ([n1, n2], link_info) in info.links.items():
            ehi_link = cls.link_info_to_ehi(link_info)
            links[(n1, n2)] = ehi_link
        return EhiNetworkInfo(links)
